"""The strategies included in this module are completed strategies that are runnable by the player
in order to do the intended """

import typing

from crapssim.bet import Come, DontCome, DontPass, Field, PassLine, Place
from crapssim.strategy.odds import (
    DontPassOddsMultiplier,
    OddsMultiplier,
    PassLineOddsMultiplier,
)
from crapssim.strategy.single_bet import (
    BetCome,
    BetDontPass,
    BetField,
    BetPassLine,
    BetPlace,
    StrategyMode,
)
from crapssim.strategy.tools import (
    AddIfNotBet,
    AddIfPointOff,
    AddIfPointOn,
    AddIfTrue,
    AggregateStrategy,
    CountStrategy,
    Player,
    RemoveByType,
    RemoveIfTrue,
    Strategy,
    WinProgression,
)


class Pass2Come(AggregateStrategy):
    """Places a PassLine bet and two Come bets. Equivalent to BetPassLine(amount) +
    CountStrategy(Come, 2, Come(amount))"""

    def __init__(self, bet_amount: float):
        """Place a PassLine bet and two Come bets of the given amount.

        Parameters
        ----------
        bet_amount
            The amount of the PassLine and Come bets.
        """
        self.bet_amount: float = float(bet_amount)
        super().__init__(
            BetPassLine(bet_amount), CountStrategy(Come, 2, Come(bet_amount))
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(amount={self.bet_amount})"


class PassLinePlace68(AggregateStrategy):
    """Bet the PassLine and Place the 6 and the 8. Equivalent to: BetPassLine(pass_line_amount) +
    BetPlace({6: six_amount, 8: eight_amount}, skip_point=skip_point)"""

    def __init__(
        self,
        pass_line_amount: float = 5,
        six_amount: float = 6,
        eight_amount: float = 6,
        skip_point: bool = True,
    ):
        """Bet the PassLine and Place the 6 and the 8. Equivalent to:
        BetPassLine(pass_line_amount) +
        BetPlace({6: six_amount, 8: eight_amount}, skip_point=skip_point)

        Parameters
        ----------
        pass_line_amount
            How much to bet on the PassLine
        six_amount
            How much to bet on the six
        eight_amount
            How much to bet on the eight
        skip_point
            If True, don't place the six or eight if that is the number of the point.
        """
        self.pass_line_amount = float(pass_line_amount)
        self.six_amount = float(six_amount)
        self.eight_amount = float(eight_amount)
        self.skip_point = skip_point

        pass_line_strategy = BetPassLine(pass_line_amount)
        six_eight_strategy = BetPlace(
            {6: six_amount, 8: eight_amount}, skip_point=skip_point
        )
        super().__init__(pass_line_strategy, six_eight_strategy)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(pass_line_amount={self.pass_line_amount}, "
            f"six_amount={self.six_amount}, eight_amount={self.eight_amount}, "
            f"skip_point={self.skip_point})"
        )


class PlaceInside(AggregateStrategy):
    """Strategy to have Place bets on all the inside (5, 6, 8, 9) numbers.
    Equivalent to BetPlace({5: x, 6: 6/5*x, 8: 6/5*x, 9: x})"""

    def __init__(self, bet_amount: typing.SupportsFloat | dict[int, float]) -> None:
        """Creates a Strategy to have Place bets on all the inside (5, 6, 8, 9) numbers.

        Parameters
        ----------
        bet_amount
            Either a dictionary of the inside numbers (5, 6, 8, 9) and the bet amounts for each,
            or a number that supports float. If its a number that supports float, the six and eight
            amounts will be the number * (6 / 5) to make the payout a whole number.
        """
        if isinstance(bet_amount, typing.SupportsFloat):
            self.bet_amount = float(bet_amount)
            six_eight_amount = bet_amount * (6 / 5)
            amount_dict = {
                5: self.bet_amount,
                6: six_eight_amount,
                8: six_eight_amount,
                9: self.bet_amount,
            }
        else:
            amount_dict = {k: float(v) for k, v in bet_amount.items()}
            self.bet_amount = amount_dict

        super().__init__(BetPlace(amount_dict, skip_point=False, skip_come=False))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(amount={self.bet_amount})"


class Place68Move59(Strategy):
    """Strategy that makes place bets on the six and eight, and then if a PassLine or Come bet with
    that point comes up, moves the place bet to 5 or 9."""

    def __init__(
        self,
        pass_come_amount: float = 5,
        six_eight_amount: float = 6,
        five_nine_amount: float = 5,
    ):
        """Makes Place bets of 6 and 8 for the six_eight_amount, then if a PassLine or Come bet
        comes with the pass_come_amount, moves those bets to the five and nine with the
        five_nine_amount.

        Parameters
        ----------
        pass_come_amount
            The amount of the PassLine and Come bets.
        six_eight_amount
            The amount of the Place6 and Place8 bets.
        five_nine_amount
            The amount of the Place5 and Place9 bets.
        """
        super().__init__()
        self.pass_come_amount = float(pass_come_amount)
        self.six_eight_amount = float(six_eight_amount)
        self.five_nine_amount = float(five_nine_amount)

    def completed(self, player: Player) -> bool:
        """The strategy is completed if the player has no bets on the table, and the players
        bankroll is too low to make any of the other bets.

        Parameters
        ----------
        player
            The Player whose bankroll and bets to check.

        Returns
        -------
        True if the strategy can't continue, otherwise False.
        """
        return (
            len(player.bets) == 0
            and player.bankroll < self.pass_come_amount
            and player.bankroll < self.six_eight_amount
            and player.bankroll < self.five_nine_amount
        )

    def get_pass_line_come_points(self, player: Player) -> list[int]:
        """Get the point number (or the table point number in the case of PassLine) for any PassLine
        or Come bets on the table.

        Parameters
        ----------
        player
            The player to check the bets for.

        Returns
        -------
        A list of integers of the points for the PassLine and Come bets.
        """
        pass_line_come_points = []
        for number in (6, 8, 9, 10):
            if (
                player.table.point.number == number
                and PassLine(self.pass_come_amount) in player.bets
            ):
                pass_line_come_points.append(number)
            elif Come(self.pass_come_amount, number) in player.bets:
                pass_line_come_points.append(number)
        return pass_line_come_points

    def update_bets(self, player: Player) -> None:
        """Do nothing if the point status is Off, otherwise go through the numbers 6, 8, 5 and 9.
        If the player has both a PassLine or Come bet for that number and a Place bet, remove the
        Place bet, otherwise move on to the next number. If the player doesn't have a PassLine or
        Come bet for that number, and the place bet for that number isn't already on the table,
        make a Place bet for that number.

        Parameters
        ----------
        player
            The player to check on bets and add bets for.
        """
        place_amounts = {
            5: self.five_nine_amount,
            6: self.six_eight_amount,
            8: self.six_eight_amount,
            9: self.five_nine_amount,
        }

        if player.table.point.status == "Off":
            return

        for number in (6, 8, 5, 9):
            bet = Place(number, place_amounts[number])

            if number in self.get_pass_line_come_points(player):
                if bet in player.bets:
                    player.remove_bet(bet)
                continue

            if len([x for x in player.bets if isinstance(x, Place)]) < 2:
                AddIfNotBet(bet).update_bets(player)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(pass_come_amount={self.pass_come_amount}, "
            f"six_eight_amount={self.six_eight_amount}, "
            f"five_nine_amount={self.five_nine_amount})"
        )


class PassLinePlace68Move59(AggregateStrategy):
    """Strategy that makes a PassLine bet, makes Place bets on the six and eight, and then moves
    them back to the 5 and 9 if the point for the PassLine (also the tables point) is six or
    eight. Equivalent to BetPassLine(pass_line_amount) + Place68Move59(pass_line_amount,
    six_eight_amount, five_nine_amount)."""

    def __init__(
        self,
        pass_line_amount: float = 5,
        six_eight_amount: float = 6,
        five_nine_amount: float = 5,
    ):
        """Place a PassLine bet, Place the six and eight, and move them to 5 9 if the point for the
        PassLine bet is a 6 or 8.

        Equivalent to BetPassLine(...) + Place68Move59(...)

        Parameters
        ----------
        pass_line_amount
            The amount of the PassLine bet.
        six_eight_amount
            The amount of the Place6 and Place8 bets.
        five_nine_amount
            The amount of the Place5 and Place9 bets.
        """
        self.pass_line_amount = float(pass_line_amount)
        self.six_eight_amount = float(six_eight_amount)
        self.five_nine_amount = float(five_nine_amount)
        super().__init__(
            BetPassLine(pass_line_amount),
            Place68Move59(pass_line_amount, six_eight_amount, five_nine_amount),
        )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(pass_line_amount={self.pass_line_amount}, "
            f"six_eight_amount={self.six_eight_amount}, "
            f"five_nine_amount={self.five_nine_amount})"
        )


class Place682Come(AggregateStrategy):
    """Strategy that bets the PassLine if the point is off and there are less than 4 bets on the
    table. If the point is On, places the 6 and 8 and if there are less than 2 Come bets on the
    table and less than 4 bets overall places a Come bet."""

    def __init__(
        self,
        pass_come_amount: float = 5,
        six_eight_amount: float = 6,
        five_nine_amount: float = 5,
        *strategies: Strategy,
    ):
        """Place the six and the eight and place two come bets moving the six and eight to five or
        nine if the Come or PassLine bets points come up to those numbers. Also, if there is a
        Place6 or Place8 bet and the point if Off make a PassLine bet.

        Parameters
        ----------
        pass_come_amount
            The amount of the Come bet.
        six_eight_amount
            The amount of the Place6 and Place8 bets.
        five_nine_amount
            The amount of the Place5 and Place9 bets.
        """
        super().__init__(*strategies)
        self.pass_come_amount = float(pass_come_amount)
        self.six_eight_amount = float(six_eight_amount)
        self.five_nine_amount = float(five_nine_amount)

    def update_bets(self, player: Player) -> None:
        """If the player has less than 2 PassLine and Come bets, make the bet (depending on whether
        the point is on or off.) If the point is on, place the 6 and 8 unless there is a PassLine or
        Come bet with those then move them to the 5 or 9.

        Parameters
        ----------
        player
            The player to check on and make the bets for.
        """

        pass_come_count = len(
            [x for x in player.bets if isinstance(x, (PassLine, Come))]
        )
        if pass_come_count < 2:
            BetPassLine(self.pass_come_amount).update_bets(player)  # if point off
            BetCome(self.pass_come_amount).update_bets(player)  # if point on

        Place68Move59(
            self.pass_come_amount, self.six_eight_amount, self.five_nine_amount
        ).update_bets(player)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(pass_come_amount={self.pass_come_amount}, "
            f"six_eight_amount={self.six_eight_amount}, "
            f"five_nine_amount={self.five_nine_amount})"
        )


class IronCross(AggregateStrategy):
    """Strategy that bets the PassLine, bets the PassLine Odds, and bets Place on the 5, 6, and 8.
    If the point is on and there is no bet on the field, place a bet on the field. Equivalent to:
    BetPassLine(...) + PassLineOddsMultiplier(2), + BetPlace({...}) + AddIfPointOn(Field(...))
    """

    def __init__(self, base_amount: float):
        """Creates the IronCross strategy based on the base_amount, using that number to determine
        the amounts for all the other numbers.

        Parameters
        ----------
        base_amount
            The base amount of the bets. This amount is used for the PassLine and Field.
            base_amount * (6/5) * 2 is used for placing the six and eight, and base amount * 2
            is used for placing the five.
        """
        self.base_amount = float(base_amount)
        place_amounts = {
            5: base_amount * 2,
            6: (6 / 5) * base_amount * 2,
            8: (6 / 5) * base_amount * 2,
        }

        super().__init__(
            BetPassLine(base_amount),
            PassLineOddsMultiplier(2),
            BetPlace(place_amounts, skip_point=True),
            AddIfPointOn(Field(base_amount)),
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(base_amount={self.base_amount})"


class HammerLock(Strategy):
    """Strategy that makes a PassLine bet and a DontPass bet when the point is off. Once the point
    is on, adds LayOdds to the DontPass bet, and Places the 6 and 8. If either of those place bets
    win, shifts the bets outside to the 5, 6, 8, and 9. If one of those wins, all place bets get
    taken down.
    """

    def __init__(self, base_amount: float):
        """Creates the HammerLock strategy with all bet amounts being created from the given
        base_amount.

        Parameters
        ----------
        base_amount
            the base amount for PassLine and DontPass Bets, and Place5 and Place9 bets. Place6 and
            Place8 starts at (6/5) * 2 * this amount and if it wins moves to (6 / 5) * this amount.
        """
        self.base_amount = float(base_amount)
        self.start_six_eight_amount = (6 / 5) * base_amount * 2
        self.end_six_eight_amount = (6 / 5) * base_amount
        self.five_nine_amount = base_amount
        self.odds_multiplier = 6

        self.place_win_count: int = 0

    def completed(self, player: Player) -> bool:
        """The strategy is completed if the player can no longer make the initial PassLine bet
        because their bankroll is too low, and they have no more bets on the table.

        Parameters
        ----------
        player

        Returns
        -------

        """
        return player.bankroll < self.base_amount and len(player.bets) == 0

    def after_roll(self, player: Player) -> None:
        """Update the place_win_count based on how many Place bets are won. If table.point.status is
        On and the dice total is 7 (meaning the shooter sevens out) reset place_win_count to 0.

        Parameters
        ----------
        player
        """
        place_bets = player.get_bets_by_type((Place,))
        winning_place_bets = [
            bet for bet in place_bets if bet.get_result(player.table).won
        ]
        self.place_win_count += len(winning_place_bets)
        if player.table.point.status == "On" and player.table.dice.total == 7:
            self.place_win_count = 0

    def update_bets(self, player: Player) -> None:
        """If the point is off bet the PassLine and DontPass line. If the point is on bet the
        Place6 and Place8 until one wins, then bet the Place 5, 6, 8, and 9. LayOdds whenever
        possible.

        Parameters
        ----------
        player
            Player to place the bets for.
        """
        if player.table.point.status == "Off":
            self.pass_and_dontpass(player)
        elif self.place_win_count == 0:
            self.place68(player)
        elif self.place_win_count == 1:
            self.place5689(player)
        elif self.place_win_count == 2:
            RemoveByType(Place).update_bets(player)
        DontPassOddsMultiplier(self.odds_multiplier).update_bets(player)

    def pass_and_dontpass(self, player: Player) -> None:
        """Update bets when point is Off: add a PassLine and a DontPass bet if they don't already exist."""
        RemoveByType(Place).update_bets(player)
        BetPassLine(self.base_amount, StrategyMode.ADD_IF_NOT_BET).update_bets(player)
        BetDontPass(self.base_amount, StrategyMode.ADD_IF_NOT_BET).update_bets(player)

    def place68(self, player: Player) -> None:
        """Update bets to Place the 6 and 8 (regardless of the point) and then lay odds on DontPass bets."""
        place_amounts = {
            6: self.start_six_eight_amount,
            8: self.start_six_eight_amount,
        }
        BetPlace(place_amounts, skip_point=False).update_bets(player)

    def place5689(self, player: Player) -> None:
        """Update bets to Place the 5, 6, 8 and 9."""
        RemoveByType(Place).update_bets(player)
        place_amounts = {
            5: self.five_nine_amount,
            6: self.end_six_eight_amount,
            8: self.end_six_eight_amount,
            9: self.five_nine_amount,
        }
        BetPlace(place_amounts, skip_point=False).update_bets(player)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(base_amount={self.base_amount})"


class Risk12(Strategy):
    """Strategy that makes a PassLine and Field bet before the point is established. Once the point
    is established, places either the 6 the 8, or both depending on if the player won enough
    pre-point to cover those bets."""

    def __init__(self) -> None:
        """Pass line and field bet before the point is established. Once the point is established
        place the 6 and 8.
        """
        super().__init__()
        self.pre_point_winnings: float = 0.0

    def completed(self, player: Player) -> bool:
        """The strategy is completed if the Player can no longer make the initial PassLine bet, and
        the player has no bets on the table.

        Parameters
        ----------
        player
            The player to check the bets and bankroll for.

        Returns
        -------
        True if the player can no longer continue the strategy, otherwise False.
        """
        return player.bankroll < 5 and len(player.bets) == 0

    def after_roll(self, player: Player) -> None:
        """Determine the pre-point winnings which is used to determine which bets to place when the
        point is on.

        Parameters
        ----------
        player
            The player to check the bets for.
        """
        bet_results = [x.get_result(player.table) for x in player.bets]
        if player.table.point.status == "Off" and any(x.won for x in bet_results):
            self.pre_point_winnings += sum(x.bankroll_change for x in bet_results)
        elif player.table.point.status == "On" and player.table.dice.total == 7:
            self.pre_point_winnings = 0

    @staticmethod
    def point_off(player: Player) -> None:
        """Place a 5 PassLine and Field bet.

        Parameters
        ----------
        player
            The player to check the bets for.
        """
        AddIfNotBet(PassLine(5)).update_bets(player)
        AddIfNotBet(Field(5)).update_bets(player)

    def point_on(self, player: Player) -> None:
        """If your winnings were enough to cover the place bets (throwing in another dollar for
        each) make the place bets.

        Parameters
        ----------
        player
            The player to place the bets for.
        """
        if self.pre_point_winnings >= 6 - 1:
            if player.table.point.number != 6:
                AddIfNotBet(Place(6, 6)).update_bets(player)
            else:
                AddIfNotBet(Place(8, 6)).update_bets(player)
        if self.pre_point_winnings >= 12 - 2:
            BetPlace({6: 6, 8: 6}).update_bets(player)

    def update_bets(self, player: Player) -> None:
        """If the point is off make a Field and PassLine bet. If the point is on
        Place the 6 and 8 if you made enough pre-point to cover the bets.

        Parameters
        ----------
        player
            The player to make the bets for.
        """
        if player.table.point.status == "Off":
            self.point_off(player)
        elif player.table.point.status == "On":
            self.point_on(player)


class Knockout(AggregateStrategy):
    """PassLine and Don't bet prior to point, 345x PassLine Odds after point.

    Equivalent to:
    BetPassLine(amount) + AddIfPointOff(DontPass(amount)) +
    PassLineOddsMultiplier({4: 3, 5: 4, 6: 5, 8: 5, 9: 4, 10: 3})
    """

    def __init__(self, base_amount: typing.SupportsFloat) -> None:
        self.base_amount = float(base_amount)
        super().__init__(
            BetPassLine(base_amount),
            AddIfPointOff(DontPass(base_amount)),
            PassLineOddsMultiplier({4: 3, 5: 4, 6: 5, 8: 5, 9: 4, 10: 3}),
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(base_amount={self.base_amount})"


class DiceDoctor(WinProgression):
    """Field progression strategy with progressive increases and decreases. Equivalent to:
    FieldWinProgression([10, 20, 15, 30, 25, 50, 35, 70, 50, 100, 75, 150])"""

    def __init__(self, base_amount: float = 10) -> None:
        """Field bet with a progression if you win of [10, 20, 15, 30, 25, 50, 35, 70, 50, 100, 75,
        150]
        """
        self.base_amount = float(base_amount)
        progression = [10, 20, 15, 30, 25, 50, 35, 70, 50, 100, 75, 150]
        multipliers = [x / 10 for x in progression]
        super().__init__(Field(self.base_amount), multipliers=multipliers)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(base_amount={self.base_amount})"


class Place68PR(Strategy):
    """Place 6 and 8 with a "Press and Regress" approach. Strategy that places the 6 and 8.
    If either of those bets win, the bet is pressed to 2 * the bet amount. If the bet is won again,
    it is reduced to the original bet amount.
    """

    def __init__(self, base_amount: float = 6) -> None:
        """If point is on place the 6 & 8 of the amount. If you win press the bet to double. If you win
        again reduce the bet back to starting amount.

        Parameters
        ----------
        base_amount
            The starting amount of bet to place.
        """
        self.base_amount = float(base_amount)
        self.starting_amount = float(base_amount)
        self.press_amount = 2 * self.starting_amount

        self.win_one_amount = base_amount * (7 / 6)
        self.win_two_amount = base_amount * 2 * (7 / 6)

        self.six_winnings = 0.0
        self.eight_winnings = 0.0

    def completed(self, player: Player) -> bool:
        """Returns True if the players bankroll is below the bet amount and the player no longer
        has bets on the table.

        Parameters
        ----------
        player

        Returns
        -------

        """
        return player.bankroll < self.starting_amount and len(player.bets) == 0

    def after_roll(self, player: Player) -> None:
        """Get the winnings on the Place 6 and 8 bets to determine whether to press or regress.

        Parameters
        ----------
        player
            The player to check the bets for.
        """
        place_bets = player.get_bets_by_type((Place,))
        place_six_bets = [x for x in place_bets if x.number == 6]
        place_six_win_amounts = [
            x.get_result(player.table).amount - x.amount
            for x in place_six_bets
            if x.get_result(player.table).won
        ]
        self.six_winnings = sum(place_six_win_amounts)
        place_eight_bets = [x for x in place_bets if x.number == 8]
        place_eight_win_amounts = [
            x.get_result(player.table).amount - x.amount
            for x in place_eight_bets
            if x.get_result(player.table).won
        ]
        self.eight_winnings = sum(place_eight_win_amounts)

    def ensure_bets_exist(self, player: Player) -> None:
        """Ensure that there is always a place 6 or place 8 bet if the point is On.

        Parameters
        ----------
        player
            The player to place the bets for.
        """
        if player.table.point.status == "Off":
            return
        for number in (6, 8):
            if (
                Place(number, self.starting_amount) not in player.bets
                and Place(number, self.press_amount) not in player.bets
            ):
                player.add_bet(Place(number, self.starting_amount))

    def press(self, player: Player) -> None:
        """Double the bet amount of the place bets.

        Parameters
        ----------
        player
            The player to make the bets for.
        """
        if self.six_winnings == self.win_one_amount:
            player.add_bet(Place(6, self.starting_amount))
        if self.eight_winnings == self.win_one_amount:
            player.add_bet(Place(8, self.starting_amount))

    def update_bets(self, player: Player) -> None:
        """Ensure that a Place6 and Place8 bet always exist for the player of base amount.
        Press the bet if you win and haven't pressed the bet yet.

        Parameters
        ----------
        player
            The player to place the bets for.
        """
        self.ensure_bets_exist(player)
        self.press(player)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(base_amount={self.base_amount})"


class Place68DontCome2Odds(AggregateStrategy):
    """Strategy that adds a DontCome bet when the point is Off, places the 6 and 8 when the point
    is On and adds 2x Odds to the DontCome bet."""

    def __init__(
        self, six_eight_amount: float = 6, dont_come_amount: float = 5
    ) -> None:
        """Place the 6 and 8 along with a Don't Come bet with 2x odds.

        Parameters
        ----------
        six_eight_amount
            The amount of the Place6 and Place8 bet.
        dont_come_amount
            The amount of the DontCome bet.
        """
        self.six_eight_amount = float(six_eight_amount)
        self.dont_come_amount = float(dont_come_amount)
        super().__init__(
            BetPlace({6: six_eight_amount, 8: six_eight_amount}, skip_point=False),
            AddIfTrue(
                DontCome(dont_come_amount),
                lambda p: len(p.get_bets_by_type((DontCome,))) == 0,
            ),
            OddsMultiplier(DontCome, 2),
        )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(six_eight_amount={self.six_eight_amount}, "
            f"dont_come_amount={self.dont_come_amount})"
        )



from crapssim import Dice


class Point:
    """
    The point on a craps table.

    Attributes
    ----------
    number : int
        The point number (in [4, 5, 6, 8, 9, 10]) is status == 'On'
    """

    def __init__(self, number: int | None = None) -> None:
        self.number: int | None = number

    @property
    def status(self) -> str:
        if self.number is None:
            return 'Off'
        else:
            return 'On'

    def __hash__(self) -> int:
        return hash(self.number)

    def __repr__(self):
        return f'Point(number={self.number})'

    def __eq__(self, other: object) -> bool:
        if isinstance(other, str):
            return self.status.lower() == other.lower() or str(self.number) == other
        elif isinstance(other, int) and other in (4, 5, 6, 8, 9, 10):
            return other == self.number
        elif isinstance(other, Point):
            return other.status == self.status and other.number == self.number
        else:
            raise NotImplementedError

    def __gt__(self, other: object) -> bool:
        if self.number is None:
            raise NotImplementedError
        if isinstance(other, str):
            return self.number > int(other)
        elif isinstance(other, int):
            return self.number > other
        elif isinstance(other, Point):
            if other.number is None:
                raise NotImplementedError
            return self.number > other.number
        else:
            raise NotImplementedError

    def __lt__(self, other: object) -> bool:
        if self.number is None:
            raise NotImplementedError
        if isinstance(other, str):
            return self.number < int(other)
        elif isinstance(other, int):
            return self.number < other
        elif isinstance(other, Point):
            if other.number is None:
                raise NotImplementedError
            return self.number < other.number
        else:
            raise NotImplementedError

    def __ge__(self, other: object) -> bool:
        if self.number is None:
            raise NotImplementedError
        return self.__eq__(other) or self.__gt__(other)

    def __le__(self, other: object) -> bool:
        if self.number is None:
            raise NotImplementedError
        return self.__eq__(other) or self.__lt__(other)

    def update(self, dice_object: Dice) -> None:
        """
        Given a Dice object update the points status and number.

        Parameters
        ----------
        dice_object : Dice
            The Dice you want to update the point with
        """
        if self.status == "Off" and dice_object.total in [4, 5, 6, 8, 9, 10]:
            self.number = dice_object.total
        elif self.status == "On" and dice_object.total in [7, self.number]:
            self.number = None


import copy
import typing
from abc import ABC, ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import Protocol, TypedDict

from crapssim.dice import Dice
from crapssim.point import Point

__all__ = [
    "BetResult",
    "Bet",
    "_WinningLosingNumbersBet",
    "_SimpleBet",
    "PassLine",
    "Come",
    "DontPass",
    "DontCome",
    "Odds",
    "Place",
    "Field",
    "CAndE",
    "Any7",
    "Two",
    "Three",
    "Yo",
    "Boxcars",
    "AnyCraps",
    "HardWay",
    "Hop",
    "Fire",
    "All",
    "Tall",
    "Small",
]
ALL_DICE_NUMBERS = {2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12}


class TableSettings(TypedDict):

    ATS_payouts: dict[str, int]  # {"all": 150, "tall": 30, "small": 30}
    field_payouts: dict[int, int]  # {2: 2, 3: 1, 4: 1, 9: 1, 10: 1, 11: 1, 12: 2}
    fire_payouts: dict[int, int]  # {4: 24, 5: 249, 6: 999}
    hop_payouts: dict[str, int]  # {"easy": 15, "hard": 30}
    max_odds: dict[int, int]  # {4: 3, 5: 4, 6: 5, 8: 5, 9: 4, 10: 3}
    max_dont_odds: dict[int, int]  # {4: 6, 5: 6, 6: 6, 8: 6, 9: 6, 10: 6}


class Table(Protocol):
    dice: Dice
    point: Point
    settings: TableSettings


class Player(Protocol):
    table: Table
    bets: list[typing.Type["Bet"]]


@dataclass(slots=True, frozen=True)
class BetResult:
    """
    Represents the outcome of a bet

    This class is used by all Bets for consistency in determining whether the
    bet won, lost or pushed. It provides properties to analyze the bet outcome
    and its impact on a bankroll.
    """

    amount: float
    """The monetary value representing the bet outcome."""
    remove: bool
    """Flag indicating whether this bet result should be removed from table."""

    @property
    def won(self) -> bool:
        """Returns True if the bet won (positive amount)."""
        return self.amount > 0

    @property
    def lost(self) -> bool:
        """Returns True if the bet lost (negative amount)."""
        return self.amount < 0

    @property
    def pushed(self) -> bool:
        """Returns True if the bet tied (zero amount)."""
        return self.amount == 0

    @property
    def bankroll_change(self) -> float:
        """Calculates the change to the bankroll (amount if bet won, zero otherwise)."""
        if self.won:
            return self.amount
        else:
            return 0


class _MetaBetABC(ABCMeta):
    # Trick to get a bet like `PassLine` to have it's repr be `crapssim.bet.PassLine`
    def __repr__(cls):
        return f"crapssim.bet.{cls.__name__}"


class Bet(ABC, metaclass=_MetaBetABC):
    """
    A generic bet for the craps table.

    The high-level class that defines most of the core bet methods.
    All bets will be a subclass of this.
    """

    def __init__(self, amount: typing.SupportsFloat):
        self.amount: float = float(amount)
        """Wagered amount for the bet."""

    @abstractmethod
    def get_result(self, table: Table) -> BetResult:
        """
        Core bet logic that determines the result.

        This determines the ultimate amount and whether the
        bet needs to be removed, which is indicated with a
        BetResult object.
        """
        pass

    def update_number(self, table: Table):
        """
        Update the bet's number, if applicable

        This method is required by Come and DontCome bets to
        update their number after the first roll. Since this
        method is used by the Table, it defaults to doing nothing
        for a generic bet.
        """
        pass

    def is_removable(self, table: Table) -> bool:
        """
        Checks whether the bet is removable. May depend on the
        table conditions (e.g. if point is On).

        Returns:
            True if the bet is removable, otherwise false.
        """
        return True

    def is_allowed(self, player: Player) -> bool:
        """
        Checks whether the bet is allowed to be placed on the given table.
        May depend on the player's bets also (e.g. for odds bets).

        Returns:
            True if the bet is allowed, otherwise false.
        """
        return True

    @property
    def _placed_key(self) -> typing.Hashable:
        return type(self)

    @property
    def _hash_key(self) -> typing.Hashable:
        return self._placed_key, self.amount

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Bet):
            return self._hash_key == other._hash_key
        raise NotImplementedError

    def __hash__(self) -> int:
        return hash(self._hash_key)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(amount={self.amount})"

    def __add__(self, other: "Bet") -> "Bet":
        if isinstance(other, typing.SupportsFloat):
            amount = self.amount - float(other)
        elif self._placed_key == other._placed_key:
            amount = self.amount + other.amount
        else:
            raise NotImplementedError
        new_bet = copy.copy(self)
        new_bet.amount = amount
        return new_bet

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other: "Bet") -> "Bet":
        if isinstance(other, typing.SupportsFloat):
            amount = self.amount - float(other)
        elif self._placed_key == other._placed_key:
            amount = self.amount - other.amount
        else:
            raise NotImplementedError
        new_bet = copy.copy(self)
        new_bet.amount = amount
        return new_bet

    def __rsub__(self, other):
        return self.__sub__(other)


class _WinningLosingNumbersBet(Bet, ABC):
    """
    A bet that has winning numbers, losing numbers, and payout ratios

    These values (possibly depending on the table) are used to
    calculate the result.
    """

    def get_result(self, table: Table) -> BetResult:
        """Core bet logic that determines the result.

        Wins are based on having dice total in the winning numbers
        (which may depend on the table), which will pay the payout_ratio
        times the bet amount plus the original bet amount back. Losses
        happen when dice total is in the losing numbers, which result
        in a loss of the original bet amount. Otherwise the bet stays
        on the table.
        """
        if table.dice.total in self.get_winning_numbers(table):
            result_amount = self.get_payout_ratio(table) * self.amount + self.amount
            should_remove = True
        elif table.dice.total in self.get_losing_numbers(table):
            result_amount = -1 * self.amount
            should_remove = True
        else:
            result_amount = 0
            should_remove = False

        return BetResult(result_amount, should_remove)

    @abstractmethod
    def get_winning_numbers(self, table: Table) -> list[int]:
        """Returns the winnings numbers, based on table features"""
        pass

    @abstractmethod
    def get_losing_numbers(self, table: Table) -> list[int]:
        """Returns the losing numbers, based on table features"""
        pass

    @abstractmethod
    def get_payout_ratio(self, table: Table) -> float:
        """Returns the payout ratio (X to 1), based on table features"""
        pass


class _SimpleBet(_WinningLosingNumbersBet, ABC):
    """
    A bet that has fixed winning and losing numbers and payout ratio

    Essentially, the numbers and payout ratio can be known
    at instantiation and don't depend on the table.
    """

    winning_numbers: list[int] = []
    """Winning numbers for the bet"""
    losing_numbers: list[int] = []
    """Losing numbers for the bet"""
    payout_ratio: int = 1
    """Payout ratio for the bet"""

    def get_winning_numbers(self, table: Table) -> list[int]:
        """Returns the winning numbers (table not used here)"""
        return self.winning_numbers

    def get_losing_numbers(self, table: Table) -> list[int]:
        """Returns the losing numbers (table not used here)"""
        return self.losing_numbers

    def get_payout_ratio(self, table: Table) -> float:
        """Returns the payout ratio (table not used here)"""
        return float(self.payout_ratio)


# Passline and related bets ---------------------------------------------------


class PassLine(_WinningLosingNumbersBet):
    """
    Pass Line bet in craps.

    A bet where the player wins if the first roll is 7 or 11,
    loses if the first roll is 2, 3, or 12, and establishes a point number
    for subsequent rolls. Once a point is set, the player wins by rolling
    the point number again before rolling a 7. Pays 1 to 1.
    """

    def get_winning_numbers(self, table: Table) -> list[int]:
        """Winnings numbers are 7, 11 before point is set,
        and the point number after point is set. Uses table
        to determine the point number and status.
        """
        if table.point.number is None:
            return [7, 11]
        return [table.point.number]

    def get_losing_numbers(self, table: Table) -> list[int]:
        """Losing numbers are 2, 3, 12 before point is set,
        and 7 after point is set. Uses table to determine the
        point number and status.
        """
        if table.point.number is None:
            return [2, 3, 12]
        return [7]

    def get_payout_ratio(self, table: Table) -> float:
        """PassLine always pays out 1:1"""
        return 1.0

    def is_removable(self, table: Table) -> bool:
        """PassLine is removable if the point is off

        Returns:
            True if the bet is removable, otherwise false.
        """
        return table.point.status == "Off"

    def is_allowed(self, player: Player) -> bool:
        """PassLine is allowed if the point if off

        Returns:
            True if the bet is allowed, otherwise false.
        """
        return player.table.point.status == "Off"


class Come(_WinningLosingNumbersBet):
    """
    Come bet in craps.

    Similar to the Pass Line bet, but can be placed after a point is established.
    The first roll determines the Come bet's point number, but also wins on 7, 11
    and loses on 2, 3, 12. The bet wins in subsequent rolls if the
    point number is rolled before a 7, and loses if a 7 is rolled before
    the point number. Pays 1 to 1.
    """

    def __init__(self, amount: typing.SupportsFloat, number: int | None = None):
        super().__init__(amount)
        possible_numbers = (4, 5, 6, 7, 8, 9, 10)
        if number in possible_numbers:
            self.number = number
        else:
            self.number = None

    def get_winning_numbers(self, table: Table) -> list[int]:
        """Winnings numbers are 7, 11 before the number is set,
        and the number after it is set. Number is stored within
        the bet.
        """
        if self.number is None:
            return [7, 11]
        return [self.number]

    def get_losing_numbers(self, table: Table) -> list[int]:
        """Losing numbers are 2, 3, 12 before the number is set,
        and 7 after it is set. Number is stored within
        the bet.
        """
        if self.number is None:
            return [2, 3, 12]
        return [7]

    def get_payout_ratio(self, table: Table) -> float:
        """Come always pays out 1:1"""
        return 1.0

    def update_number(self, table: Table):
        """
        Update the bet's number to the first number rolled if it's in (4, 5, 6, 8, 9, 10).
        """
        possible_numbers = (4, 5, 6, 8, 9, 10)
        if self.number is None and table.dice.total in possible_numbers:
            self.number = table.dice.total

    def is_removable(self, table: Table) -> bool:
        """Come bet is removable is it's number has not been established yet (first roll).

        Returns:
            True if the bet is removable, otherwise false.
        """
        return self.number is None

    def is_allowed(self, player: Player) -> bool:
        """Come bet is only allowed if the table's point is on (accessed via player).

        Returns:
            True if the bet is allowed, otherwise false.
        """
        return player.table.point.status == "On"

    @property
    def _placed_key(self) -> typing.Hashable:
        return type(self), self.number

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(amount={self.amount}, number={self.number})"


class DontPass(_WinningLosingNumbersBet):
    """
    Don't Pass bet in craps.

    The opposite of the Pass Line bet. The player wins if the first roll is 2 or 3,
    pushes on 12, and loses if the first roll is 7 or 11. After a point is
    established, the player wins by rolling a 7 before the point number. Bet pays 1 to 1.

    Note that a push will keep the bet active and not result in any change to bankroll.
    """

    def get_winning_numbers(self, table: Table) -> list[int]:
        """Winnings numbers are 2 or 3 before point is set,
        and 7 after point is set. Uses table to determine the point
        number and status.
        """
        if table.point.number is None:
            return [2, 3]
        return [7]

    def get_losing_numbers(self, table: Table) -> list[int]:
        """Losing numbers are 7 or 11 before point is set,
        and table point number after point is set. Uses table to determine the
        point number and status.
        """
        if table.point.number is None:
            return [7, 11]
        return [table.point.number]

    def get_payout_ratio(self, table: Table) -> float:
        """Don't pass always pays out 1:1"""
        return 1.0

    def is_allowed(self, player: Player) -> bool:
        """Don't Pass is allowed if the point if off.

        Returns:
            True if the bet is allowed, otherwise false.
        """
        return player.table.point.status == "Off"


class DontCome(_WinningLosingNumbersBet):
    """
    Don't Come bet in craps.

    Similar to the Don't Pass bet, but can be placed after a point is
    established, but also wins on 2, 3, pushes on 12, and loses on 7 or 11.
    The first roll determines the Don't Come bet's number. The bet wins in
    subsequent rolls if a 7 is rolled before the point number, and loses if
    the number is rolled before a 7. Pays 1 to 1.
    """

    def __init__(self, amount: typing.SupportsFloat, number: int | None = None):
        super().__init__(amount)
        possible_numbers = (4, 5, 6, 7, 8, 9, 10)
        if number in possible_numbers:
            self.number = number
        else:
            self.number = None

    def get_winning_numbers(self, table: Table) -> list[int]:
        if self.number is None:
            return [2, 3]
        return [7]

    def get_losing_numbers(self, table: Table) -> list[int]:
        if self.number is None:
            return [7, 11]
        return [self.number]

    def get_payout_ratio(self, table: Table) -> float:
        """Don't Come always pays out 1:1"""
        return 1.0

    def update_number(self, table: Table):
        possible_numbers = (4, 5, 6, 7, 8, 9, 10)
        if self.number is None and table.dice.total in possible_numbers:
            self.number = table.dice.total

    def is_allowed(self, player: Player) -> bool:
        """Don't Come is only allowed if the table's point is on (accessed via player).

        Returns:
            True if the bet is allowed, otherwise false.
        """
        return player.table.point.status == "On"

    @property
    def _placed_key(self) -> typing.Hashable:
        return type(self), self.number

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(amount={self.amount}, number={self.number})"


# Odds bets -------------------------------------------------------------------


class Odds(_WinningLosingNumbersBet):
    """
    Odds bet (for PassLine, DontPass, Come, or Dontcome) in craps.

    A supplementary bet placed behind Pass Line, Don't Pass, Come, or Don't Come bets.
    Offers true odds payouts, meaning the house has no edge. The payout varies
    depending on the point number and whether it's a "light side" (Pass/Come)
    or "dark side" (Don't Pass/Don't Come) bet.
    """

    def __init__(
        self,
        base_type: typing.Type[PassLine | DontPass | Come | DontCome],
        number: int,
        amount: float,
    ):
        super().__init__(amount)
        self.base_type = base_type
        self.number = number

    @property
    def light_side(self) -> bool:
        return issubclass(self.base_type, (PassLine, Come))

    @property
    def dark_side(self) -> bool:
        return issubclass(self.base_type, (DontPass, DontCome))

    def get_winning_numbers(self, table: Table) -> list[int]:
        if self.light_side:
            return [self.number]
        elif self.dark_side:
            return [7]

    def get_losing_numbers(self, table: Table) -> list[int]:
        if self.light_side:
            return [7]
        elif self.dark_side:
            return [self.number]

    def get_payout_ratio(self, table: Table) -> float:
        light_ratios = {4: 2, 5: 3 / 2, 6: 6 / 5, 8: 6 / 5, 9: 3 / 2, 10: 2}
        dark_ratios = {n: 1 / x for n, x in light_ratios.items()}

        if self.light_side:
            return light_ratios[self.number]
        elif self.dark_side:
            return dark_ratios[self.number]

    def is_allowed(self, player: Player) -> bool:
        """Odds are allowed if they do not exceed the table maximums.

        Returns:
            True if the bet is allowed, otherwise false.
        """
        max_bet = self.get_max_odds(player.table) * self.base_amount(player)
        return self.amount <= max_bet

    def get_max_odds(self, table: Table) -> float:
        if self.light_side:
            return table.settings["max_odds"][self.number]
        elif self.dark_side:
            return table.settings["max_dont_odds"][self.number]
        else:
            raise NotImplementedError

    def base_amount(self, player: Player):
        base_bets = [
            x
            for x in player.bets
            if isinstance(x, self.base_type)
            and x.get_winning_numbers(player.table)
            == self.get_winning_numbers(player.table)
        ]
        return sum(x.amount for x in base_bets)

    @property
    def _placed_key(self) -> typing.Hashable:
        return type(self), self.base_type, self.number

    def __repr__(self):
        return (
            f"Odds(base_type={self.base_type}, number={self.number}, "
            f"amount={self.amount})"
        )


# Place bets ------------------------------------------------------------------


class Place(_SimpleBet):
    """
    Place bet (on 4, 5, 6, 8, 9, or 10) in craps.

    A bet on a specific number (4, 5, 6, 8, 9, or 10) being rolled before a 7.
    Each number has a different payout ratio reflecting its probability of being rolled.
    Remains active until the number or a 7 is rolled.
    """

    payout_ratios = {4: 9 / 5, 5: 7 / 5, 6: 7 / 6, 8: 7 / 6, 9: 7 / 5, 10: 9 / 5}
    """Stores the place bet payouts: 9 to 5 on (4, 10), 7 to 5 on (5, 9), and 7 to 6 on (6, 8)."""
    losing_numbers: list[int] = [7]

    def __init__(self, number: int, amount: typing.SupportsFloat):
        super().__init__(amount)
        self.number = number
        """The placed number, which determines payout ratio"""
        self.payout_ratio = self.payout_ratios[number]
        self.winning_numbers = [number]

    @property
    def _placed_key(self) -> typing.Hashable:
        return type(self), self.number

    def __repr__(self) -> str:
        return f"Place({self.winning_numbers[0]}, amount={self.amount})"


# _WinningLosingNumbersBets with variable payouts -----------------------------------------------------------------


class Field(_WinningLosingNumbersBet):
    """
    Field bet in craps.

    A one-roll bet that wins if the next roll is 2, 3, 4, 9, 10, 11, or 12.
    Loses if 5, 6, 7, or 8 are rolled. Offers variable payouts for specific numbers
    as defined in the table settings (:func:`~crapssim.table.TableSettings`,
    "field_payouts":, which default to 2 to 1 for (2, 12) and 1 to 1 otherwise.
    """

    winning_numbers = [2, 3, 4, 9, 10, 11, 12]
    """Field wins on 2, 3, 4, 9, 10, 11, or 12"""
    losing_numbers = [5, 6, 7, 8]
    """Field loses on 5, 6, 7, or 8"""

    def get_winning_numbers(self, table: Table) -> list[int]:
        """Returns the winning numbers (table not used here)"""
        return self.winning_numbers

    def get_losing_numbers(self, table: Table) -> list[int]:
        """Returns the losing numbers (table not used here)"""
        return self.losing_numbers

    def get_payout_ratio(self, table: Table) -> float:
        """Returns the payout ratio (X to 1) based on table settings
        (:func:`~crapssim.table.TableSettings`, "field_payouts":
        """
        if table.dice.total in table.settings["field_payouts"]:
            return float(table.settings["field_payouts"][table.dice.total])
        return 0.0


class CAndE(_WinningLosingNumbersBet):
    """
    Craps and Eleven (C & E) bet in craps.

    A one-roll bet that wins if the next roll is 2, 3, 11, or 12.
    Offers different payout ratios for different winning numbers:
    - 3 to 1 for 2, 3, and 12
    - 7 to 1 for 11
    Loses on all other numbers.
    """

    winning_numbers: list[int] = [2, 3, 11, 12]
    """Winning numbers are (2, 3, 11, 12)."""
    losing_numbers: list[int] = list(ALL_DICE_NUMBERS - {2, 3, 11, 12})
    """Losing numbers are anything besides (2, 3, 11, 12)."""

    def get_winning_numbers(self, table: Table) -> list[int]:
        """Returns the winning numbers (table not used here)"""
        return self.winning_numbers

    def get_losing_numbers(self, table: Table) -> list[int]:
        """Returns the losing numbers (table not used here)"""
        return self.losing_numbers

    def get_payout_ratio(self, table: Table) -> float:
        """C & E pays out 3 to 1 for (2, 3, 12) and 7 to 1 for (11)."""
        if table.dice.total in [2, 3, 12]:
            return 3.0
        elif table.dice.total in [11]:
            return 7.0
        else:
            raise NotImplementedError


# Simple bets in the middle of the table --------------------------------------


class Any7(_SimpleBet):
    """
    Any 7 bet (also known as Big Red) in craps.

    A one-roll bet that wins only if the next roll is 7.
    Offers a 4 to 1 payout and loses on all other numbers.
    """

    winning_numbers: list[int] = [7]
    losing_numbers: list[int] = list(ALL_DICE_NUMBERS - {7})
    """Losing number is anything except 7."""
    payout_ratio: int = 4


class Two(_SimpleBet):
    """
    Two (Snake Eyes) bet in craps.

    A one-roll bet that wins only if the next roll is 2.
    Offers a 30 to 1 payout and loses on all other numbers.
    """

    winning_numbers: list[int] = [2]
    losing_numbers: list[int] = list(ALL_DICE_NUMBERS - {2})
    """Losing number is anything except 2."""
    payout_ratio: int = 30


class Three(_SimpleBet):
    """
    Three bet in craps.

    A one-roll bet that wins only if the next roll is 3.
    Offers a 15 to 1 payout and loses on all other numbers.
    """

    winning_numbers: list[int] = [3]
    losing_numbers: list[int] = list(ALL_DICE_NUMBERS - {3})
    """Losing number is anything except 3."""
    payout_ratio: int = 15


class Yo(_SimpleBet):
    """
    Yo (Eleven) bet in craps.

    A one-roll bet that wins only if the next roll is 11.
    Offers a 15 to 1 payout and loses on all other numbers.
    """

    winning_numbers: list[int] = [11]
    losing_numbers: list[int] = list(ALL_DICE_NUMBERS - {11})
    """Losing number is anything except 11."""
    payout_ratio: int = 15


class Boxcars(_SimpleBet):
    """
    Boxcars (Midnight) bet in craps.

    A one-roll bet that wins only if the next roll is 12.
    Offers a 30 to 1 payout and loses on all other numbers.
    """

    winning_numbers: list[int] = [12]
    losing_numbers: list[int] = list(ALL_DICE_NUMBERS - {12})
    """Losing number is anything except 12."""
    payout_ratio: int = 30


class AnyCraps(_SimpleBet):
    """
    Any Craps bet in craps.

    A one-roll bet that wins if the next roll is 2, 3, or 12.
    Offers a 7 to 1 payout and loses on all other numbers.
    """

    winning_numbers: list[int] = [2, 3, 12]
    losing_numbers: list[int] = list(ALL_DICE_NUMBERS - {2, 3, 12})
    """Losing number is anything except (2, 3, 12)."""
    payout_ratio: int = 7


# HardWay Bets ----------------------------------------------------------------


class HardWay(Bet):
    """
    Hard Way bet (on 4, 6, 8, or 10) in craps.

    A bet on rolling a specific even number (4, 6, 8, or 10)
    with both dice showing the same value (e.g., two 2s for a hard 4).
    Wins if the number is rolled in a "hard" way before either a 7 or
    the number is rolled in a "soft" way.
    """

    payout_ratios = {4: 7, 6: 9, 8: 9, 10: 7}
    """Payout ratios vary: 7 to 1 for hard 4 or 10, 9 to 1 for hard 6 or 8."""

    def __init__(self, number: int, amount: typing.SupportsFloat) -> None:
        super().__init__(amount)
        self.number: int = number
        self.payout_ratio: float = self.payout_ratios[number]

    def get_result(self, table: Table) -> BetResult:
        if table.dice.result == self.winning_result:
            result_amount = self.payout_ratio * self.amount + self.amount
            should_remove = True
        elif table.dice.total in (7, self.number):
            result_amount = -1 * self.amount
            should_remove = True
        else:
            result_amount = 0
            should_remove = False
        return BetResult(result_amount, should_remove)

    @property
    def winning_result(self) -> tuple[int, int]:
        """Returns the dice result that wins, e.g. (2, 2) for Hard 4."""
        return (int(self.number / 2), int(self.number / 2))

    @property
    def _placed_key(self) -> typing.Hashable:
        return type(self), self.number

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.number}, amount={self.amount})"


# Hop bets -------------------------------------------------------------------


class Hop(Bet):
    """
    Hop bet in craps.

    A one-roll bet on a specific dice combination.
    Can be an "easy" hop (different values on each die) or a "hard" hop
    (same value on both dice).

    Payouts differ based on whether the hop is easy or hard, and are
    set in the table settings (:func:`~crapssim.table.TableSettings`,
    "hop_payouts")
    - Easy hop: standard payout (default 15 to 1)
    - Hard hop: higher payout (default 30 to 1)
    """

    def __init__(self, result: tuple[int, int], amount: typing.SupportsFloat) -> None:
        super().__init__(amount)
        self.result: tuple[int, int] = tuple(sorted(result))

    def get_result(self, table: Table) -> BetResult:
        if table.dice.result in self.winning_results:
            result_amount = self.payout_ratio(table) * self.amount + self.amount
            should_remove = True
        else:
            result_amount = -1 * self.amount
            should_remove = True
        return BetResult(result_amount, should_remove)

    @property
    def is_easy(self) -> bool:
        return self.result[0] != self.result[1]

    @property
    def winning_results(self) -> list[tuple[int, int]]:
        if self.is_easy:
            return [self.result, self.result[::-1]]
        else:
            return [self.result]

    def payout_ratio(self, table: Table) -> int:
        payout_type = "easy" if self.is_easy else "hard"
        return table.settings["hop_payouts"][payout_type]

    @property
    def _placed_key(self) -> typing.Hashable:
        return type(self), self.result

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.result}, amount={self.amount})"


# Fire bet -------------------------------------------------------------------


class Fire(Bet):
    """
    Fire bet in craps.

    A progressive bet that tracks points made during a shooter's turn.
    Wins with increasing payouts based on the number of unique point
    numbers made before a 7 is rolled.

    Payout escalates as more points are made:
    - Specific payout ratios depend on table settings (:func:`~crapssim.table.TableSettings`,
    "fire_payouts"), default is 24 to 1 for four points, 249 to 1 for five points, and
    999 to 1 for all six points.
    - Automatically ends when all 6 points are made or a 7 is rolled while the point is On.
    """

    def __init__(self, amount: float):
        super().__init__(amount)
        self.points_made: set[int] = set()
        self.ended: bool = False

    def get_result(self, table: Table) -> BetResult:

        if table.point.status == "Off":
            return BetResult(amount=0, remove=False)

        if table.dice.total == table.point.number:
            self.points_made.add(table.point.number)

        # Fire pays out on 7 when enough points made
        # Fire pays out automatically when all 6 points are made
        n_points_made = len(self.points_made)
        ended = table.dice.total == 7 or len(self.points_made) == 6

        if ended and n_points_made in table.settings["fire_payouts"]:
            payout_ratio = table.settings["fire_payouts"][n_points_made]
            result_amount = payout_ratio * self.amount + self.amount
        elif ended and n_points_made not in table.settings["fire_payouts"]:
            result_amount = -1 * self.amount
        else:
            result_amount = 0

        return BetResult(result_amount, remove=ended)

    def is_removable(self, table: Table) -> bool:
        """Fire bet is removable only if there is a new shooter.

        Returns:
            True if the bet is removable, otherwise false.
        """
        return table.new_shooter

    def is_allowed(self, player: Player) -> bool:
        """Fire bet is allowed if there is a new shooter.

        Returns:
            True if the bet is allowed, otherwise false.
        """
        return player.table.new_shooter


# All-tall-small bets -------------------------------------------------------


class _ATSBet(Bet):
    """Class representing ATS (All, Tall, Small) bets, not a usable bet by itself."""

    numbers: list[int] = []
    type: str = "_ATSBet"

    def __init__(self, amount: float):
        super().__init__(amount)
        self.rolled_numbers: set[int] = set()

    def get_result(self, table: Table) -> BetResult:

        if table.dice.total in self.numbers:
            self.rolled_numbers.add(table.dice.total)

        if self.numbers == list(self.rolled_numbers):
            payout_ratio = table.settings["ATS_payouts"][self.type]
            result_amount = payout_ratio * self.amount + self.amount
            should_remove = True
        elif table.dice.total == 7:
            result_amount = -1 * self.amount
            should_remove = True
        else:
            result_amount = 0
            should_remove = False

        return BetResult(result_amount, should_remove)

    def is_removable(self, table: Table) -> bool:
        """All/Tall/Small bets are removable only if there is a new shooter.

        Returns:
            True if the bet is removable, otherwise false.
        """
        return table.new_shooter

    def is_allowed(self, player: Player) -> bool:
        """All/Tall/Small bets are allowed if there is a new shooter.

        Returns:
            True if the bet is allowed, otherwise false.
        """
        return player.table.new_shooter


class All(_ATSBet):
    """
    All bet (part of All/Tall/Small bets) in craps.

    Wins when 2, 3, 4, 5, 6, 8, 9, 10, 11, and 12 all roll
    before a 7 rolls. Loses immediately if a 7 is rolled (including come-out
    sevens). Payout ratios are determined by the :func:`~crapssim.table.TableSettings`
    (["ATS_payouts"]["all"]), which defaults to 150 to 1.
    """

    type: str = "all"
    numbers: list[int] = [2, 3, 4, 5, 6, 8, 9, 10, 11, 12]


class Tall(_ATSBet):
    """
    Tall bet (part of All/Tall/Small bets) in craps.

    Wins when 8, 9, 10, 11, and 12 all roll
    before a 7 rolls. Loses immediately if a 7 is rolled (including come-out
    sevens). Payout ratios are determined by the :func:`~crapssim.table.TableSettings`
    (["ATS_payouts"]["tall"]), which defaults to 30 to 1.
    """

    type: str = "tall"
    numbers: list[int] = [8, 9, 10, 11, 12]


class Small(_ATSBet):
    """
    Small bet (part of All/Tall/Small bets) in craps.

    Wins when 2, 3, 4, 5, and 6 all roll
    before a 7 rolls. Loses immediately if a 7 is rolled (including come-out
    sevens). Payout ratios are determined by the :func:`~crapssim.table.TableSettings`
    (["ATS_payouts"]["small"]), which defaults to 30.
    """

    type: str = "small"
    numbers: list[int] = [2, 3, 4, 5, 6]



"""
The dice are used by the craps Table for keeping track of the latest roll
and the total number of rolls so far. The dice object is mostly handled 
internally, but advanced users may access it (through the Table, as table.dice)
for new bets or strategies as needed. 
"""

import typing

import numpy as np


class Dice:
    """
    Simulate the rolling of a dice.

    Args:
        seed (int): The seed passed to the random number generator.
    """

    def __init__(self, seed=None) -> None:
        self._result: typing.Iterable[int] | None = None
        self.n_rolls: int = 0
        """Number of rolls for the dice"""
        self.rng: typing.Generator = np.random.default_rng(seed)
        """Random number generated used when rolling"""

    @property
    def total(self) -> int:
        """Sum of dice outcome, e.g. 8 for (2, 6)"""
        if self._result is not None:
            return sum(self.result)

    @property
    def result(self) -> tuple[int, int]:
        """Most recent outcome of the roll of two dice, e.g. (2, 6)"""
        if self._result is not None:
            return tuple(self._result)

    @result.setter
    def result(self, value: typing.Iterable[int]) -> tuple[int, int]:
        # Allows setting of result, used for some tests, but not recommended
        # NOTE: no checking is done here, so use with caution
        # NOTE: this does not increment the number of rolls
        self._result = value

    def roll(self) -> None:
        """
        Randomly roll the dice

        The randomness of the dice is based on numpy.random,
        which uses the PCG-64 pseudo-random number generation
        (see numpy.random.PCG64`).
        """
        self.n_rolls += 1
        self._result = self.rng.integers(1, 7, size=2).tolist()

    def fixed_roll(self, outcome: typing.Iterable[int]) -> None:
        """
        Roll the dice with a specified outcome

        Args:
            outcome: The desired dice result to roll
        """
        self.n_rolls += 1
        self._result = outcome



from itertools import combinations

import pytest

from crapssim import Dice, Table
from crapssim.bet import (
    All,
    Any7,
    AnyCraps,
    Boxcars,
    CAndE,
    Come,
    DontCome,
    DontPass,
    Field,
    Fire,
    HardWay,
    Odds,
    PassLine,
    Place,
    Small,
    Tall,
    Three,
    Two,
    Yo,
)
from crapssim.point import Point
from crapssim.strategy.tools import NullStrategy
from crapssim.table import TableUpdate

ALL_BETS = [
    PassLine(5),
    Come(5),
    Odds(PassLine, 4, 5),
    Odds(PassLine, 5, 5),
    Odds(PassLine, 6, 5),
    Odds(PassLine, 8, 5),
    Odds(PassLine, 9, 5),
    Odds(PassLine, 10, 5),
    Odds(Come, 4, 5),
    Odds(Come, 5, 5),
    Odds(Come, 6, 5),
    Odds(Come, 8, 5),
    Odds(Come, 9, 5),
    Odds(Come, 10, 5),
    Place(4, 5),
    Place(5, 5),
    Place(6, 5),
    Place(8, 5),
    Place(9, 5),
    Place(10, 5),
    Field(5),
    DontPass(5),
    DontCome(5),
    Odds(DontPass, 4, 5),
    Odds(DontPass, 5, 5),
    Odds(DontPass, 6, 5),
    Odds(DontPass, 8, 5),
    Odds(DontPass, 9, 5),
    Odds(DontPass, 10, 5),
    Odds(DontCome, 4, 5),
    Odds(DontCome, 5, 5),
    Odds(DontCome, 6, 5),
    Odds(DontCome, 8, 5),
    Odds(DontCome, 9, 5),
    Odds(DontCome, 10, 5),
    Any7(5),
    Two(5),
    Three(5),
    Yo(5),
    Boxcars(5),
    AnyCraps(5),
    CAndE(5),
    HardWay(4, 5),
    HardWay(6, 5),
    HardWay(8, 5),
    HardWay(10, 5),
    Fire(5),
    All(5),
    Small(5),
    Tall(5),
]


@pytest.mark.parametrize("bet_one, bet_two", [(x, x) for x in ALL_BETS])
def test_bet_equality2(bet_one, bet_two):
    assert bet_one == bet_two


@pytest.mark.parametrize("bet_one, bet_two", [x for x in combinations(ALL_BETS, r=2)])
def test_bet_type_inequality(bet_one, bet_two):
    assert bet_one != bet_two


@pytest.mark.parametrize(
    "bet_one, bet_two",
    [
        (PassLine(10), PassLine(15)),
        (Come(25), Come(10)),
        (Odds(PassLine, 4, 5), Odds(PassLine, 4, 20)),
        (Odds(PassLine, 5, 10), Odds(PassLine, 5, 25)),
        (Odds(PassLine, 6, 25), Odds(PassLine, 6, 10)),
        (Odds(PassLine, 8, 20), Odds(PassLine, 8, 30)),
        (Odds(PassLine, 9, 15), Odds(PassLine, 9, 5)),
        (Odds(PassLine, 10, 20), Odds(PassLine, 10, 5)),
        (Place(4, 10), Place(4, 30)),
        (Place(5, 30), Place(5, 10)),
        (Place(6, 25), Place(6, 20)),
        (Place(8, 30), Place(8, 15)),
        (Place(9, 15), Place(9, 25)),
        (Place(10, 5), Place(10, 10)),
        (Field(30), Field(5)),
        (DontPass(20), DontPass(5)),
        (DontCome(15), DontCome(25)),
        (Odds(DontPass, 4, 10), Odds(DontPass, 4, 25)),
        (Odds(DontPass, 5, 10), Odds(DontPass, 5, 15)),
        (Odds(DontPass, 6, 30), Odds(DontPass, 6, 5)),
        (Odds(DontPass, 8, 30), Odds(DontPass, 8, 10)),
        (Odds(DontPass, 9, 5), Odds(DontPass, 9, 10)),
        (Odds(DontPass, 10, 10), Odds(DontPass, 10, 30)),
        (Any7(30), Any7(25)),
        (Two(20), Two(25)),
        (Three(10), Three(25)),
        (Yo(30), Yo(10)),
        (Boxcars(30), Boxcars(10)),
        (AnyCraps(25), AnyCraps(15)),
        (CAndE(5), CAndE(25)),
        (HardWay(4, 25), HardWay(4, 10)),
        (HardWay(6, 15), HardWay(6, 25)),
        (HardWay(8, 15), HardWay(8, 5)),
        (HardWay(10, 20), HardWay(10, 5)),
        (Fire(20), Fire(30)),
        (All(2), All(7)),
        (Small(5), Small(10)),
        (Tall(5), Tall(30)),
    ],
)
def test_bet_amount_inequality(bet_one, bet_two):
    assert bet_one != bet_two


@pytest.mark.parametrize(
    "bet",
    [
        PassLine(5),
        Odds(PassLine, 4, 5),
        Odds(PassLine, 5, 5),
        Odds(PassLine, 6, 5),
        Odds(PassLine, 8, 5),
        Odds(PassLine, 9, 5),
        Odds(PassLine, 10, 5),
        Place(4, 5),
        Place(5, 5),
        Place(6, 6),
        Place(8, 8),
        Place(9, 9),
        Place(10, 10),
        Odds(DontPass, 4, 5),
        Odds(DontPass, 5, 5),
        Odds(DontPass, 6, 5),
        Odds(DontPass, 8, 5),
        Odds(DontPass, 9, 5),
        Odds(DontPass, 10, 5),
        Field(5),
        DontPass(5),
        DontCome(5),
        Any7(5),
        Two(5),
        Three(5),
        Yo(5),
        Boxcars(5),
        AnyCraps(5),
        HardWay(4, 5),
        HardWay(6, 5),
        HardWay(8, 5),
        HardWay(10, 5),
    ],
)
def test_is_removable_table_point_off(bet):
    table = Table()
    assert bet.is_removable(table) is True


@pytest.mark.parametrize(
    "bet",
    [
        Odds(PassLine, 4, 5),
        Odds(PassLine, 5, 5),
        Odds(PassLine, 6, 5),
        Odds(PassLine, 8, 5),
        Odds(PassLine, 9, 5),
        Odds(PassLine, 10, 5),
        Place(4, 5),
        Place(5, 5),
        Place(6, 6),
        Place(8, 8),
        Place(9, 9),
        Place(10, 10),
        Odds(DontPass, 4, 5),
        Odds(DontPass, 5, 5),
        Odds(DontPass, 6, 5),
        Odds(DontPass, 8, 5),
        Odds(DontPass, 9, 5),
        Odds(DontPass, 10, 5),
        Field(5),
        DontPass(5),
        DontCome(5),
        Any7(5),
        Two(5),
        Three(5),
        Yo(5),
        Boxcars(5),
        AnyCraps(5),
        HardWay(4, 5),
        HardWay(6, 5),
        HardWay(8, 5),
        HardWay(10, 5),
    ],
)
def test_is_removable_table_point_on(bet):
    table = Table()
    table.point.number = 6
    assert bet.is_removable(table) is True


@pytest.mark.parametrize(
    "bet, new_shooter, is_removable",
    [
        (Fire(5), True, True),
        (Fire(5), False, False),
        (All(5), True, True),
        (All(5), False, False),
        (Tall(5), True, True),
        (Tall(5), False, False),
        (Small(5), True, True),
        (Small(5), False, False),
    ],
)
def test_bet_is_removable_new_shooter(bet, new_shooter, is_removable):
    table = Table()
    table.new_shooter = new_shooter

    assert bet.is_removable(table) == is_removable


@pytest.mark.parametrize(
    "dice1, dice2, correct_ratio",
    [(1, 1, 2), (1, 2, 1), (2, 2, 1), (5, 4, 1), (5, 5, 1), (6, 5, 1), (6, 6, 2)],
)
def test_get_field_default_table_payout_ratio(dice1, dice2, correct_ratio):
    table = Table()
    table.dice.fixed_roll((dice1, dice2))
    assert Field(5).get_payout_ratio(table) == correct_ratio


@pytest.mark.parametrize(
    "dice1, dice2, correct_ratio",
    [(1, 1, 2), (1, 2, 14), (2, 2, 14000), (5, 4, 1), (5, 5, 1), (6, 5, 1), (6, 6, 3)],
)
def test_get_field_non_default_table_payout_ratio(dice1, dice2, correct_ratio):
    table = Table()
    table.settings["field_payouts"].update({3: 14, 12: 3, 4: 14000})
    table.dice.fixed_roll((dice1, dice2))
    assert Field(5).get_payout_ratio(table) == correct_ratio


@pytest.mark.parametrize(
    "points_made, correct_ratio",
    [
        ({4, 5, 6, 9}, 24),
        ({4, 5, 6, 9, 10}, 249),
        ({4, 5, 6, 8, 9, 10}, 999),
    ],
)
def test_get_fire_default_table_payout_ratio(points_made, correct_ratio):
    table = Table()
    bet = Fire(1)
    table.point.number = 8
    table.dice.result = [3, 4]  # 7-out
    bet.points_made = points_made

    ratio = (bet.get_result(table).amount - bet.amount) / bet.amount
    assert ratio == correct_ratio


@pytest.mark.parametrize(
    "points_made, correct_ratio",
    [
        ({4, 5, 6}, 6),
        ({4, 5, 6, 9}, 9),
        ({4, 5, 6, 9, 10}, 69),
        ({4, 5, 6, 8, 9, 10}, 420),
    ],
)
def test_get_fire_non_default_table_payout_ratio(points_made, correct_ratio):
    table = Table()
    table.settings["fire_payouts"] = {3: 6, 4: 9, 5: 69, 6: 420}
    bet = Fire(1)
    table.point.number = 8
    table.dice.result = [3, 4]  # 7-out
    bet.points_made = points_made

    ratio = (bet.get_result(table).amount - bet.amount) / bet.amount
    assert ratio == correct_ratio


# fmt: off
@pytest.mark.parametrize(
    'rolls, correct_bankroll_change, correct_value_change, correct_exists', 
    [
        (
            [(6, 1)], 
            -1, 0, True
        ),
        (
            [(2, 2), (3, 1), (4, 3), (6, 6)], 
            -1, 0, True
        ),
        (
            [(2, 2), (4, 3)], 
            -1, -1, False
        ),
        (
            [(2, 2), (2, 2), (3, 3), (3, 3), (4, 3), (4, 4), (4, 4), (5, 5), 
             (5, 5)],
            -1, 0, True,
        ),
        (
            [(2, 2), (2, 2), (3, 3), (3, 3), (4, 3), (4, 4), (4, 4), (5, 5), 
             (5, 5), (5, 5), (5, 5)],
            -1, 0, True,
        ),
        (
            [(2, 2), (2, 2), (3, 3), (3, 3), (4, 3), (4, 4), (4, 4), (5, 5), 
             (5, 5), (5, 5), (3, 4)],
            24, 24, False
        ),
        (
            [(2, 2), (2, 2), (3, 3), (3, 3), (4, 3), (4, 4), (4, 4), (5, 5), 
             (5, 5), (2, 3), (2, 3)],
            -1, 0, True,
        ),
        (
            [(2, 2), (2, 2), (3, 3), (3, 3), (4, 3), (4, 4), (4, 4), (5, 5), 
             (5, 5), (2, 3), (2, 3), (5, 4)],
            -1, 0, True
        ),
        (
            [(2, 2), (2, 2), (3, 3), (3, 3), (4, 3), (4, 4), (4, 4), (5, 5), 
             (5, 5), (2, 3), (2, 3), (4, 4), (3, 4)],
            249, 249, False
        ),
        (
            [(2, 2), (2, 2), (2, 3), (3, 2), (3, 3), (3, 3), (4, 4), (4, 4), 
             (4, 5), (5, 4), (5, 5), (5, 5)],
            999, 999, False
        ),
    ]
)
# fmt: on
def test_fire_on_table(
    rolls: list[tuple[int]],
    correct_bankroll_change: float,
    correct_value_change: float,
    correct_exists: bool,
):

    table = Table()
    start_bankroll = 100
    table.add_player(bankroll=start_bankroll, strategy=NullStrategy())
    player = table.players[0]
    player.add_bet(Fire(1))

    table.fixed_run(rolls, verbose=True)

    bankroll_change = player.bankroll - start_bankroll
    value_change = player.bankroll + player.total_bet_amount - start_bankroll
    exists = player.has_bets(Fire)

    assert (bankroll_change, value_change, exists) == (
        correct_bankroll_change,
        correct_value_change,
        correct_exists,
    )


@pytest.mark.parametrize(
    "bet, point_number, is_allowed",
    [
        (PassLine(5), None, True),
        (PassLine(5), 6, False),
        (Come(5), None, False),
        (Come(5), 6, True),
        (DontPass(5), None, True),
        (DontPass(5), 4, False),
        (DontCome(5), None, False),
        (DontCome(5), 8, True),
        (Field(5), None, True),
        (Field(5), 4, True),
    ],
)
def test_bet_is_allowed_point(bet, point_number, is_allowed):
    table = Table()
    table.add_player()
    dice = Dice()
    if point_number is None:
        dice.result = None
    else:
        dice.result = [point_number // 2, point_number - point_number // 2]
    # dice.total = point_number

    point = Point()
    point.update(dice)

    table.point = point

    assert bet.is_allowed(player=table.players[0]) == is_allowed


@pytest.mark.parametrize(
    "bet, new_shooter, is_allowed",
    [
        (Field(5), True, True),
        (Field(5), False, True),
        (Fire(5), True, True),
        (Fire(5), False, False),
        (All(5), True, True),
        (All(5), False, False),
        (Tall(5), True, True),
        (Tall(5), False, False),
        (Small(5), True, True),
        (Small(5), False, False),
    ],
)
def test_bet_is_allowed_new_shooter(bet, new_shooter, is_allowed):
    table = Table()
    table.add_player()

    table.new_shooter = new_shooter

    assert bet.is_allowed(player=table.players[0]) == is_allowed


@pytest.mark.parametrize(
    "bet",
    [
        PassLine(5),
        Place(4, 5),
        Place(5, 5),
        Place(6, 5),
        Place(8, 5),
        Place(9, 5),
        Place(10, 5),
        DontPass(5),
        Field(5),
        Any7(5),
        Two(5),
        Three(5),
        Yo(5),
        Boxcars(5),
        AnyCraps(5),
        HardWay(4, 5),
        HardWay(6, 5),
        HardWay(8, 5),
        HardWay(10, 5),
    ],
)
def test_bets_always_is_allowed_point_off(bet):
    table = Table()
    table.add_player()
    assert bet.is_allowed(table.players[0])


@pytest.mark.parametrize(
    "bet",
    [
        Come(5),
        Place(4, 5),
        Place(5, 5),
        Place(6, 5),
        Place(8, 5),
        Place(9, 5),
        Place(10, 5),
        DontCome(5),
        Field(5),
        Any7(5),
        Two(5),
        Three(5),
        Yo(5),
        Boxcars(5),
        AnyCraps(5),
        HardWay(4, 5),
        HardWay(6, 5),
        HardWay(8, 5),
        HardWay(10, 5),
    ],
)
def test_bets_always_is_allowed_point_on(bet):
    table = Table()
    table.point.number = 10
    table.add_player()
    assert bet.is_allowed(table.players[0])


# fmt: off
@pytest.mark.parametrize('rolls, correct_bankroll_change, correct_value_change, correct_exists', [
    (
        [(2, 2)], 
        -1, 0, True
    ),
    (
        [(10, 1), (10, 2), (7, 2), (5, 5), (2, 6), (1, 1), 
         (1, 2), (2, 2), (2, 3), (3, 3)], 
        150, 150, False
    ),
    (
        [(10, 1), (10, 2), (7, 2), (5, 5), (2, 6), (1, 1), 
         (1, 2), (2, 2), (2, 3)], 
        -1, 0, True
    ),
    (
        [(1, 1), (1, 2), (2, 2), (2, 3), (3, 4)], 
        -1, -1, False
    )
])
# fmt: on
def test_all_on_table(
    rolls: list[tuple[int]],
    correct_bankroll_change: float,
    correct_value_change: float,
    correct_exists: bool,
):

    table = Table()
    start_bankroll = 100
    table.add_player(bankroll=start_bankroll, strategy=NullStrategy())
    player = table.players[0]
    player.add_bet(All(1))

    table.fixed_run(rolls, verbose=True)

    bankroll_change = player.bankroll - start_bankroll
    value_change = player.bankroll + player.total_bet_amount - start_bankroll
    exists = player.has_bets(All)

    assert (bankroll_change, value_change, exists) == (
        correct_bankroll_change,
        correct_value_change,
        correct_exists,
    )


# fmt: off
@pytest.mark.parametrize('rolls, correct_bankroll_change, correct_value_change, correct_exists', [
    (
        [(5, 6)], 
        10, 10, False
    ),
    (
        [(3, 3)], 
        -10, 0, True
    ),
])
# fmt: on
def test_passline_on_table(
    rolls: list[tuple[int]],
    correct_bankroll_change: float,
    correct_value_change: float,
    correct_exists: bool,
):

    table = Table()
    start_bankroll = 100
    table.add_player(bankroll=start_bankroll, strategy=NullStrategy())
    player = table.players[0]
    player.add_bet(PassLine(10))

    table.fixed_run(rolls, verbose=True)

    bankroll_change = player.bankroll - start_bankroll
    value_change = player.bankroll + player.total_bet_amount - start_bankroll
    exists = player.has_bets(PassLine)

    assert (bankroll_change, value_change, exists) == (
        correct_bankroll_change,
        correct_value_change,
        correct_exists,
    )


# fmt: off
@pytest.mark.parametrize('rolls, correct_bankroll_change, correct_value_change, correct_exists', [
    (
        [(2, 2)], 
        -1, 0, True
    ),
    (
        [(1, 1), (1, 2), (2, 2), (2, 3), (3, 3)], 
        -1, 0, True
    ),
    (
        [(10, 1), (10, 2), (7, 2), (5, 5), (2, 6)], 
        30, 30, False
    ),
    (
        [(10, 1), (10, 2), (7, 2), (5, 5), (3, 4)], 
        -1, -1, False
    ),
    (
        [(10, 1), (10, 2), (7, 2), (5, 5)], 
        -1, 0, True
    ),
])
# fmt: on
def test_tall_on_table(
    rolls: list[tuple[int]],
    correct_bankroll_change: float,
    correct_value_change: float,
    correct_exists: bool,
):

    table = Table()
    start_bankroll = 100
    table.add_player(bankroll=start_bankroll, strategy=NullStrategy())
    player = table.players[0]
    player.add_bet(Tall(1))

    table.fixed_run(rolls, verbose=True)

    bankroll_change = player.bankroll - start_bankroll
    value_change = player.bankroll + player.total_bet_amount - start_bankroll
    exists = player.has_bets(Tall)

    assert (bankroll_change, value_change, exists) == (
        correct_bankroll_change,
        correct_value_change,
        correct_exists,
    )


# fmt: off
@pytest.mark.parametrize('rolls, correct_bankroll_change, correct_value_change, correct_exists', [
    (
        [(2, 2)], 
        -1, 0, True
    ),
    (
        [(1, 1), (1, 2), (2, 2), (2, 3), (3, 3)], 
        30, 30, False
    ),
    (
        [(10, 1), (10, 2), (7, 2), (5, 5), (2, 6)], 
        -1, 0, True
    ),
    (
        [(1, 1), (1, 2), (2, 2), (2, 3), (3, 4)], 
        -1, -1, False
    ),
    (
        [(1, 1), (1, 2), (2, 2), (2, 3)], 
        -1, 0, True
    ),
])
# fmt: on
def test_small_on_table(
    rolls: list[tuple[int]],
    correct_bankroll_change: float,
    correct_value_change: float,
    correct_exists: bool,
):

    table = Table()
    start_bankroll = 100
    table.add_player(bankroll=start_bankroll, strategy=NullStrategy())
    player = table.players[0]
    player.add_bet(Small(1))

    table.fixed_run(rolls, verbose=True)

    bankroll_change = player.bankroll - start_bankroll
    value_change = player.bankroll + player.total_bet_amount - start_bankroll
    exists = player.has_bets(Small)

    assert (bankroll_change, value_change, exists) == (
        correct_bankroll_change,
        correct_value_change,
        correct_exists,
    )


@pytest.mark.parametrize(
    "ATS_payouts, bet, rolled_numbers, correct_ratio",
    [
        (
            {"all": 150, "tall": 30, "small": 30},
            All(1),
            [2, 3, 4, 5, 6, 8, 9, 10, 11, 12],
            150,
        ),
        (
            {"all": 150, "tall": 30, "small": 30},
            All(5),
            [2, 3, 4, 5, 6, 8, 9, 10, 11, 12],
            150,
        ),
        (
            {"all": 150, "tall": 30, "small": 30},
            Tall(1),
            [8, 9, 10, 11, 12],
            30,
        ),
        (
            {"all": 150, "tall": 30, "small": 30},
            Small(5),
            [2, 3, 4, 5, 6],
            30,
        ),
        (
            {"all": 174, "tall": 34, "small": 34},
            All(1),
            [2, 3, 4, 5, 6, 8, 9, 10, 11, 12],
            174,
        ),
        (
            {"all": 174, "tall": 34, "small": 34},
            Tall(5),
            [8, 9, 10, 11, 12],
            34,
        ),
        (
            {"all": 174, "tall": 34, "small": 34},
            Small(1),
            [2, 3, 4, 5, 6],
            34,
        ),
    ],
)
def test_all_tall_small_table_payout_ratio(
    ATS_payouts, bet, rolled_numbers, correct_ratio
):
    table = Table()
    table.settings["ATS_payouts"] = ATS_payouts
    bet.rolled_numbers = set(rolled_numbers)

    ratio = (bet.get_result(table).amount - bet.amount) / bet.amount
    assert ratio == correct_ratio



import pytest

from crapssim.dice import Dice


@pytest.fixture
def d1():
    return Dice()


def test_no_rolls(d1):
    assert d1.n_rolls == 0


def test_one_roll(d1):
    d1.roll()
    assert d1.n_rolls == 1


def test_many_roll(d1):
    d1.roll()
    d1.roll()
    d1.roll()
    d1.roll()
    assert d1.n_rolls == 4


@pytest.mark.parametrize(
    "roll, total",
    [
        ([3, 4], 7),
        ([1, 1], 2),
        ([4, 3], 7),
        ([5, 6], 11),
    ],
)
def test_fixed_roll(d1, roll, total):
    d1.fixed_roll(roll)
    assert d1.result == tuple(roll)
    assert d1.total == total


@pytest.mark.parametrize("seed", [8, 15, 21234, 0])
def test_roll_seed_idential(seed):
    d1 = Dice(seed)
    d2 = Dice(seed)

    d1.roll()
    d2.roll()
    assert d1.result == d2.result
    assert d1.total == d2.total



import pytest

from crapssim import Table
from crapssim.bet import Come, DontCome, DontPass, Field, Fire, Odds, PassLine, Place
from crapssim.strategy import (
    AddIfTrue,
    BetDontPass,
    BetPassLine,
    BetPlace,
    DontPassOddsMultiplier,
    PassLineOddsMultiplier,
)
from crapssim.strategy.examples import (
    DiceDoctor,
    HammerLock,
    IronCross,
    Knockout,
    Pass2Come,
    PassLinePlace68,
    PassLinePlace68Move59,
    Place68DontCome2Odds,
    Place68PR,
    Place682Come,
    PlaceInside,
    Risk12,
)
from crapssim.strategy.single_bet import BetFire
from crapssim.strategy.tools import AddIfNewShooter, AddIfTrue, WinProgression
from crapssim.table import TableUpdate


@pytest.mark.parametrize(
    ["strategy", "rolls", "correct_bets"],
    [
        (BetPassLine(bet_amount=5), [], [PassLine(amount=5.0)]),
        (BetPassLine(bet_amount=5), [(4, 4)], [PassLine(amount=5.0)]),
        (
            BetPassLine(bet_amount=5) + PassLineOddsMultiplier(odds_multiplier=1),
            [],
            [PassLine(amount=5.0)],
        ),
        (
            BetPassLine(bet_amount=5) + PassLineOddsMultiplier(odds_multiplier=1),
            [(4, 4)],
            [Odds(PassLine, 8, amount=5.0), PassLine(amount=5.0)],
        ),
        (
            BetPassLine(bet_amount=5) + PassLineOddsMultiplier(odds_multiplier=1),
            [(4, 4), (3, 3)],
            [Odds(PassLine, 8, amount=5.0), PassLine(amount=5.0)],
        ),
        (
            BetPassLine(bet_amount=5)
            + PassLineOddsMultiplier(
                odds_multiplier={4: 3, 5: 4, 6: 5, 8: 5, 9: 4, 10: 3}
            ),
            [],
            [PassLine(amount=5.0)],
        ),
        (
            BetPassLine(bet_amount=5)
            + PassLineOddsMultiplier(
                odds_multiplier={4: 3, 5: 4, 6: 5, 8: 5, 9: 4, 10: 3}
            ),
            [(6, 4)],
            [PassLine(amount=5.0), Odds(PassLine, 10, amount=15.0)],
        ),
        (
            BetPassLine(bet_amount=5) + PassLineOddsMultiplier(odds_multiplier=2),
            [(2, 2)],
            [PassLine(amount=5.0), Odds(PassLine, 4, amount=10.0)],
        ),
        (
            BetPassLine(bet_amount=5)
            + PassLineOddsMultiplier(
                odds_multiplier={4: 3, 5: 4, 6: 5, 8: 5, 9: 4, 10: 3}
            ),
            [(3, 4), (3, 3)],
            [PassLine(amount=5.0), Odds(PassLine, 6, amount=25.0)],
        ),
        (Pass2Come(bet_amount=5), [], [PassLine(amount=5.0)]),
        (Pass2Come(bet_amount=5), [(4, 5)], [Come(amount=5.0), PassLine(amount=5.0)]),
        (
            Pass2Come(bet_amount=5),
            [(4, 5), (5, 5)],
            [PassLine(amount=5.0), Come(amount=5.0, number=10), Come(amount=5.0)],
        ),
        (
            Pass2Come(bet_amount=5),
            [(4, 5), (5, 5), (3, 3)],
            [
                PassLine(amount=5.0),
                Come(amount=5.0, number=10),
                Come(amount=5.0, number=6),
            ],
        ),
        (BetPlace(place_bet_amounts={4: 5}, skip_point=True), [], []),
        (
            BetPlace(place_bet_amounts={5: 5}, skip_point=True),
            [(3, 3)],
            [Place(number=5, amount=5.0)],
        ),
        (BetPlace(place_bet_amounts={5: 5}, skip_point=True), [(3, 2)], []),
        (
            BetPlace(place_bet_amounts={5: 5}, skip_point=False),
            [(3, 2)],
            [Place(number=5, amount=5.0)],
        ),
        (
            PassLinePlace68(
                pass_line_amount=5, six_amount=6, eight_amount=6, skip_point=True
            ),
            [(4, 5)],
            [
                Place(number=8, amount=6.0),
                PassLine(amount=5.0),
                Place(number=6, amount=6.0),
            ],
        ),
        (
            PassLinePlace68(
                pass_line_amount=5, six_amount=6, eight_amount=6, skip_point=True
            ),
            [(2, 4)],
            [Place(number=8, amount=6.0), PassLine(amount=5.0)],
        ),
        (BetDontPass(bet_amount=5), [], [DontPass(amount=5.0)]),
        (
            BetDontPass(bet_amount=5) + DontPassOddsMultiplier(odds_multiplier=1),
            [],
            [DontPass(amount=5.0)],
        ),
        (
            BetDontPass(bet_amount=5) + DontPassOddsMultiplier(odds_multiplier=6),
            [(3, 3)],
            [Odds(DontPass, 6, amount=30.0), DontPass(amount=5.0)],
        ),
        (
            PassLinePlace68Move59(
                pass_line_amount=5, six_eight_amount=6, five_nine_amount=5
            ),
            [],
            [PassLine(amount=5.0)],
        ),
        (
            PassLinePlace68Move59(
                pass_line_amount=5, six_eight_amount=6, five_nine_amount=5
            ),
            [(3, 3)],
            [
                Place(number=8, amount=6.0),
                Place(number=5, amount=5.0),
                PassLine(amount=5.0),
            ],
        ),
        (
            PassLinePlace68Move59(
                pass_line_amount=5, six_eight_amount=6, five_nine_amount=5
            ),
            [(3, 3), (4, 4)],
            [
                Place(number=8, amount=6.0),
                Place(number=5, amount=5.0),
                PassLine(amount=5.0),
            ],
        ),
        (
            Place682Come(pass_come_amount=5, six_eight_amount=6, five_nine_amount=5),
            [],
            [PassLine(amount=5.0)],
        ),
        (
            Place682Come(pass_come_amount=5, six_eight_amount=6, five_nine_amount=5),
            [(3, 3)],
            [
                PassLine(5),
                Place(number=8, amount=6.0),
                Place(number=5, amount=5),
                Come(amount=5.0),
            ],
        ),
        (
            Place682Come(pass_come_amount=5, six_eight_amount=6, five_nine_amount=5),
            [(3, 3), (3, 6)],
            [
                Place(number=8, amount=6.0),
                Come(amount=5.0, number=9),
                PassLine(amount=5.0),
                Place(number=5, amount=5),
            ],
        ),
        (
            Place682Come(pass_come_amount=5, six_eight_amount=6, five_nine_amount=5),
            [(3, 3), (4, 4)],
            [
                Place(number=9, amount=5.0),
                Place(number=5, amount=5.0),
                PassLine(amount=5.0),
                Come(amount=5.0, number=8),
            ],
        ),
        (IronCross(base_amount=5), [], [PassLine(amount=5.0)]),
        (
            IronCross(base_amount=5),
            [(4, 4)],
            [
                PassLine(amount=5.0),
                Place(number=5, amount=10.0),
                Place(number=6, amount=12.0),
                Field(amount=5.0),
                Odds(PassLine, 8, amount=10.0),
            ],
        ),
        (HammerLock(base_amount=5), [], [DontPass(amount=5.0), PassLine(amount=5.0)]),
        (
            HammerLock(base_amount=5),
            [(3, 3)],
            [
                Odds(DontPass, 6, amount=30.0),
                PassLine(amount=5.0),
                Place(number=8, amount=12.0),
                DontPass(amount=5.0),
                Place(number=6, amount=12.0),
            ],
        ),
        (
            HammerLock(base_amount=5),
            [(3, 3), (4, 4)],
            [
                Place(number=8, amount=6.0),
                Place(number=9, amount=5.0),
                Odds(DontPass, 6, amount=30.0),
                PassLine(amount=5.0),
                Place(number=6, amount=6.0),
                DontPass(amount=5.0),
                Place(number=5, amount=5.0),
            ],
        ),
        (
            AddIfTrue(
                bet=PassLine(amount=5.0), key=lambda p: p.table.point.status == "Off"
            ),
            [],
            [PassLine(amount=5.0)],
        ),
        (Risk12(), [], [PassLine(amount=5.0), Field(amount=5.0)]),
        (
            Risk12(),
            [(1, 3)],
            [
                Place(number=8, amount=6.0),
                PassLine(amount=5.0),
                Place(number=6, amount=6.0),
            ],
        ),
        (
            Risk12(),
            [(5, 6), (2, 3)],
            [
                Place(number=8, amount=6.0),
                PassLine(amount=5.0),
                Place(number=6, amount=6.0),
            ],
        ),
        (Knockout(base_amount=5), [], [DontPass(amount=5.0), PassLine(amount=5.0)]),
        (
            Knockout(base_amount=5),
            [(4, 2)],
            [
                DontPass(amount=5.0),
                PassLine(amount=5.0),
                Odds(PassLine, 6, amount=25.0),
            ],
        ),
        (DiceDoctor(), [], [Field(amount=10.0)]),
        (DiceDoctor(), [(1, 1)], [Field(amount=20.0)]),
        (DiceDoctor(), [(1, 1), (5, 6)], [Field(amount=15.0)]),
        (DiceDoctor(), [(1, 1), (5, 6), (5, 5)], [Field(amount=30.0)]),
        (WinProgression(Place(6, 12), [1, 2, 3]), [], [Place(6, amount=12.0)]),
        (WinProgression(Place(6, 12), [1, 2, 3]), [(3, 3)], [Place(6, amount=24.0)]),
        (
            WinProgression(Place(6, 12), [1, 2, 3]),
            [(3, 3), (2, 4)],
            [Place(6, amount=36.0)],
        ),
        (
            WinProgression(Place(6, 12), [1, 2, 3]),
            [(3, 3), (2, 4), (3, 3)],
            [Place(6, amount=36.0)],
        ),
        (Place68DontCome2Odds(six_eight_amount=6, dont_come_amount=5), [], []),
        (
            Place68DontCome2Odds(six_eight_amount=6, dont_come_amount=5),
            [(4, 4)],
            [
                Place(number=8, amount=6.0),
                Place(number=6, amount=6.0),
                DontCome(amount=5.0),
            ],
        ),
        (
            Place68DontCome2Odds(six_eight_amount=6, dont_come_amount=5),
            [(4, 4), (2, 2)],
            [
                Place(number=8, amount=6.0),
                Odds(DontCome, 4, amount=10.0),
                Place(number=6, amount=6.0),
                DontCome(amount=5.0, number=4),
            ],
        ),
        (Place68PR(base_amount=6), [], []),
        (
            Place68PR(base_amount=6),
            [(4, 4)],
            [Place(number=8, amount=6.0), Place(number=6, amount=6.0)],
        ),
        (
            Place68PR(base_amount=6),
            [(2, 2), (4, 4)],
            [Place(number=8, amount=12.0), Place(number=6, amount=6.0)],
        ),
        (
            Place68PR(base_amount=6),
            [(2, 2), (4, 4), (4, 4)],
            [Place(number=8, amount=6.0), Place(number=6, amount=6.0)],
        ),
        (
            Place68PR(base_amount=6),
            [(2, 2), (4, 4), (4, 4), (4, 4)],
            [Place(number=8, amount=12.0), Place(number=6, amount=6.0)],
        ),
        (
            Place68PR(base_amount=6),
            [(2, 2), (3, 3), (3, 3), (3, 3)],
            [Place(number=8, amount=6.0), Place(number=6, amount=12.0)],
        ),
    ],
)
def test_strategies_compare_bets(
    strategy, rolls: list[tuple[int, int]], correct_bets: {(str, str, float)}
):
    table = Table()
    table.add_player(strategy=strategy)
    table.fixed_run(rolls, verbose=False)
    TableUpdate().run_strategies(table)

    bets = table.players[0].bets

    assert set(bets) == set(correct_bets)


def test_strategies_in_simulation_persistent_features():

    bankroll = 100
    strategies = {"Fire 1": AddIfNewShooter(Fire(1)), "Fire 2": BetFire(1)}

    # Simulation 1, five fire points hit
    outcomes = [
        (2, 2),
        (2, 2),
        (2, 3),
        (2, 3),
        (3, 3),
        (3, 3),
        (4, 4),
        (4, 4),
        (4, 5),
        (4, 5),
        (4, 4),  # set the last point
        (1, 6),  # seven-out
    ]
    table = Table()
    for s in strategies:
        table.add_player(bankroll, strategy=strategies[s], name=s)
    table.fixed_run(outcomes, verbose=False)
    for p in table.players:
        print(f"{p.name}, {p.bankroll}, {bankroll}, {table.dice.n_rolls}")

    assert p.bankroll == bankroll + table.settings["fire_payouts"][5]

    # Simulation 2, no fire points hit
    outcomes = [
        (2, 2),  # set a point
        (1, 6),  # seven-out
    ]
    table = Table()
    for s in strategies:
        table.add_player(bankroll, strategy=strategies[s], name=s)
    table.fixed_run(outcomes, verbose=False)
    for p in table.players:
        print(f"{p.name}, {p.bankroll}, {bankroll}, {table.dice.n_rolls}")

    assert p.bankroll == bankroll - 1


def test_placeinside_with_betpointon():

    bankroll = 100
    strategy = PlaceInside(10)

    # Simulation 1, point hit, then come-out seven, nothing should happen
    outcomes = [
        (2, 2),
        (2, 2),
        (1, 6),  # come-out seven, should not lose
    ]
    table = Table()
    table.add_player(bankroll, strategy)

    table.fixed_run(outcomes, verbose=False)
    for p in table.players:
        print(f"{p.name}, {p.bankroll}, {bankroll}, {table.dice.n_rolls}")

    assert p.total_player_cash == bankroll
    assert p.bets == []

    # Simulation 2, point hit, then come-out seven, reset point and win one
    outcomes = [
        (2, 2),
        (2, 2),
        (1, 6),  # come-out seven, should not lose
        (3, 3),
        (4, 4),  # 8 wins
        (1, 2),
    ]
    table = Table()
    table.add_player(bankroll, strategy)

    table.fixed_run(outcomes, verbose=True)
    for p in table.players:
        print(f"{p.name}, {p.bankroll}, {bankroll}, {table.dice.n_rolls}")

    assert p.total_player_cash == bankroll + 14
    assert len(p.bets) == 4



import copy
import typing

from crapssim.dice import Dice

from .bet import Bet, BetResult
from .point import Point
from .strategy import BetPassLine, Strategy

__all__ = ["TableUpdate", "TableSettings", "Table", "Player"]


class TableUpdate:
    """Object for processing a table after the dice has been rolled."""

    def run(
        self,
        table: "Table",
        dice_outcome: typing.Iterable[int] | None = None,
        verbose: bool = False,
    ):
        """Run through the roll logic of the table."""
        self.run_strategies(table, verbose)
        self.print_player_summary(table, verbose)
        self.before_roll(table)
        self.update_table_stats(table)
        self.roll(table, dice_outcome, verbose)
        self.after_roll(table)
        self.update_bets(table, verbose)
        self.set_new_shooter(table)
        self.update_numbers(table, verbose)

    @staticmethod
    def run_strategies(table: "Table", verbose=False):
        for player in table.players:
            player.strategy.update_bets(player)

    @staticmethod
    def print_player_summary(table: "Table", verbose=False):
        for player in table.players:
            if verbose:
                print(
                    f"{player.name}: Bankroll={player.bankroll}, "
                    f"Bet amount={player.total_bet_amount}, Bets={player.bets}"
                )

    @staticmethod
    def before_roll(table: "Table"):
        table.last_roll = table.dice.total

    @staticmethod
    def update_table_stats(table: "Table"):
        table.pass_rolls += 1
        if table.point == "On" and (
            table.dice.total == 7 or table.dice.total == table.point.number
        ):
            table.pass_rolls = 0

    @staticmethod
    def roll(
        table: "Table",
        fixed_outcome: typing.Iterable[int] | None = None,
        verbose: bool = False,
    ):
        if fixed_outcome is not None:
            table.dice.fixed_roll(fixed_outcome)
        else:
            table.dice.roll()
        if verbose:
            print("")
            print("Dice out!")
            print(f"Shooter rolled {table.dice.total} {table.dice.result}")

    @staticmethod
    def after_roll(table: "Table"):
        for player in table.players:
            player.strategy.after_roll(player)

    @staticmethod
    def update_bets(table: "Table", verbose=False):
        for player in table.players:
            player.update_bet(verbose=verbose)

    @staticmethod
    def set_new_shooter(table: "Table"):
        if table.point == "On" and table.dice.total == 7:
            table.new_shooter = True
            table.n_shooters += 1
        else:
            table.new_shooter = False

    @staticmethod
    def update_numbers(table: "Table", verbose: bool):
        "For Come and DontCome bets that 'move' to their number"
        for player, bet in table.yield_player_bets():
            bet.update_number(table)
        table.point.update(table.dice)

        if verbose:
            print(f"Point is {table.point.status} ({table.point.number})")


class TableSettings(typing.TypedDict):
    """
    Table settings including payouts and max odds.

    This controls the payouts for the ATS (All, Tall, Small), Field,
    Fire, and Hop bets. This also controls the maximum allowable odds
    for the table (both for light-side and dark-side bets).
    """

    ATS_payouts: dict[str, int]  # {"all": 150, "tall": 30, "small": 30}
    field_payouts: dict[int, int]  # {2: 2, 3: 1, 4: 1, 9: 1, 10: 1, 11: 1, 12: 2}
    fire_payouts: dict[int, int]  # {4: 24, 5: 249, 6: 999}
    hop_payouts: dict[str, int]  # {"easy": 15, "hard": 30}
    max_odds: dict[int, int]  # {4: 3, 5: 4, 6: 5, 8: 5, 9: 4, 10: 3}
    max_dont_odds: dict[int, int]  # {4: 6, 5: 6, 6: 6, 8: 6, 9: 6, 10: 6}


class Table:
    """
    Craps Table that contains Dice, Players, the Players' bets, and updates
    them accordingly.  Main method is run() which should simulate a craps
    table until a specified number of rolls plays out or all players run out
    of money.

    Attributes
    ----------
    players : list
        List of player objects at the table
    point : string
        The point for the table.  It is either "Off" when point is off or "On"
        when point is on.
    dice : Dice
        Dice for the table
    settings : dice[str, list[int]]
        Field payouts for the table
    pass_rolls : int
        Number of rolls for the current pass
    last_roll : int
        Total of the last roll for the table
    n_shooters : int
        How many shooters the table has had.
    new_shooter : bool
        Returns True if the previous shooters roll just ended and the next shooter hasn't shot.
    """

    def __init__(self, seed: int | None = None) -> None:
        self.players: list[Player] = []
        self.point: Point = Point()
        self.seed = seed
        self.dice: Dice = Dice(self.seed)
        self.settings: TableSettings = {
            "ATS_payouts": {"all": 150, "tall": 30, "small": 30},
            "field_payouts": {2: 2, 3: 1, 4: 1, 9: 1, 10: 1, 11: 1, 12: 2},
            "fire_payouts": {4: 24, 5: 249, 6: 999},
            "hop_payouts": {"easy": 15, "hard": 30},
            "max_odds": {4: 3, 5: 4, 6: 5, 8: 5, 9: 4, 10: 3},
            "max_dont_odds": {4: 6, 5: 6, 6: 6, 8: 6, 9: 6, 10: 6},
        }
        self.pass_rolls: int = 0
        self.last_roll: int | None = None
        self.n_shooters: int = 1
        self.new_shooter: bool = True

    def yield_player_bets(self) -> typing.Generator[tuple["Player", "Bet"], None, None]:
        for player in self.players:
            for bet in player.bets:
                yield player, bet

    def add_player(
        self,
        bankroll: typing.SupportsFloat = 100,
        strategy: Strategy = BetPassLine(5),
        name: str = None,
    ) -> None:
        """Add player object to the table

        Parameters
        ----------
        bankroll
            The players bankroll, defaults to 100.
        strategy
            The players strategy, defaults to passline.
        name
            The players name, if None defaults to "Player x" with x being the current number
            of players starting with 0 (ex. Player 0, Player 1, Player 2).

        """
        if name is None:
            name = f"Player {len(self.players)}"
        self.players.append(
            Player(table=self, bankroll=bankroll, bet_strategy=strategy, name=name)
        )

    def _setup_run(self, verbose: bool) -> None:
        """
        Setup the table to run and ensure that there is at least one player.

        Parameters
        ----------
        verbose
            If True prints a welcome message and the initial players.
        """
        if verbose and self.dice.n_rolls == 0:
            print("Welcome to the Craps Table!")
        self.ensure_one_player()
        if verbose and self.dice.n_rolls == 0:
            for player in self.players:
                print(
                    f"{player.name}: Strategy={player.strategy}, "
                    f"Bankroll={player.bankroll}"
                )
            print("")
            print("")

    def run(
        self,
        max_rolls: int,
        max_shooter: float | int = float("inf"),
        verbose: bool = True,
        runout: bool = False,
    ) -> None:
        """
        Runs the craps table until a stopping condition is met.

        Parameters
        ----------
        max_shooter : float | int
            Maximum number of shooters to run for
        max_rolls : int
            Maximum number of rolls to run for
        verbose : bool
            If true, print results from table during each roll
        runout : bool
            If true, continue past max_rolls until player has no more bets on the table
        """

        self._setup_run(verbose)
        n_rolls_start = self.dice.n_rolls
        # logic needs to count starting run as 0 shooters, not easy to set new_shooter in better way
        n_shooter_start = self.n_shooters if self.n_shooters != 1 else 0

        continue_rolling = True
        while continue_rolling:
            TableUpdate().run(self, verbose=verbose)
            continue_rolling = self.should_keep_rolling(
                max_rolls + n_rolls_start, max_shooter + n_shooter_start, runout
            )
            if not continue_rolling:
                self.n_shooters -= 1  # count was added but this shooter never rolled
                TableUpdate().print_player_summary(self, verbose=verbose)

    def fixed_run(
        self, dice_outcomes: typing.Iterable[typing.Iterable], verbose: bool = False
    ) -> None:
        """
        Give a series of fixed dice outcome and run as if that is what was rolled.

        Parameters
        ----------
        dice_outcomes
            Iterable with two integers representing the dice faces.
        verbose
            If true, print results from table during each roll
        """
        self._setup_run(verbose=verbose)

        for dice_outcome in dice_outcomes:
            TableUpdate().run(self, dice_outcome, verbose=verbose)

    def should_keep_rolling(
        self, max_rolls: float | int, max_shooter: float | int, runout: bool
    ) -> bool:
        """
        Determines whether the program should keep running or not.

        Parameters
        ----------
        max_rolls
            Maximum number of rolls to run for
        max_shooter
            Maximum number of shooters to run for
        runout
            If true, continue past max_rolls until player has no more bets on the table

        Returns
        -------
        If True, the program should continue running. If False the program should stop running.
        """
        if runout:
            return (
                self.dice.n_rolls < max_rolls
                and self.n_shooters <= max_shooter
                and not any(x.strategy.completed(x) for x in self.players)
            ) or self.player_has_bets
        else:
            return (
                self.dice.n_rolls < max_rolls
                and self.n_shooters <= max_shooter
                and not any(x.strategy.completed(x) for x in self.players)
            )

    def ensure_one_player(self) -> None:
        """Make sure there is at least one player at the table"""
        if len(self.players) == 0:
            self.add_player()

    @property
    def player_has_bets(self) -> bool:
        """
        Returns whether any of the players on the table have any active bets.

        Returns
        -------
        True if any of the players have bets on the table, otherwise False.
        """
        return sum([len(p.bets) for p in self.players]) > 0

    @property
    def total_player_cash(self) -> float:
        """
        Returns the total sum of all players total_bet_amounts and bankroll.

        Returns
        -------
        The total sum of all players total_bet_amounts and bankroll.
        """
        return sum([p.total_player_cash for p in self.players])


class Player:
    """
    Player standing at the craps table

    Parameters
    ----------
    bankroll : typing.SupportsFloat
        Starting amount of cash for the player
    bet_strategy : function(table, player, unit=5)
        A function that implements a particular betting strategy.  See betting_strategies.py
    name : string, default = "Player"
        Name of the player

    Attributes
    ----------
    bankroll : typing.SupportsFloat
        Current amount of cash for the player
    name : str
        Name of the player
    bet_strategy :
        A function that implements a particular betting strategy. See betting_strategies.py.
    bets : list
        List of betting objects for the player
    """

    def __init__(
        self,
        table: Table,
        bankroll: typing.SupportsFloat,
        bet_strategy: Strategy = BetPassLine(5),
        name: str = "Player",
    ):
        self.bankroll: float = float(bankroll)
        self.strategy: Strategy = copy.deepcopy(bet_strategy)
        self.name: str = name
        self.bets: list[Bet] = []
        self._table: Table = table

    @property
    def total_bet_amount(self) -> float:
        return sum(x.amount for x in self.bets)

    @property
    def total_player_cash(self) -> float:
        return self.bankroll + self.total_bet_amount

    @property
    def table(self) -> Table:
        return self._table

    def add_bet(self, bet: Bet) -> None:
        existing_bets: list[Bet] = self.already_placed_bets(bet)
        new_bet = sum(existing_bets + [bet])
        amount_available_to_bet = self.bankroll + sum(x.amount for x in existing_bets)

        if new_bet.is_allowed(self) and new_bet.amount <= amount_available_to_bet:
            for bet in existing_bets:
                self.bets.remove(bet)
            self.bankroll -= bet.amount
            self.bets.append(new_bet)

    def already_placed_bets(self, bet: Bet) -> list[Bet]:
        """
        Returns the bets a player has matching the placed key

        Notably, bets like Place(4, 1.0) will not match to Place(6, 1.0).
        """
        return [x for x in self.bets if x._placed_key == bet._placed_key]

    def already_placed(self, bet: Bet) -> bool:
        return len(self.already_placed_bets(bet)) > 0

    def get_bets_by_type(
        self, bet_type: typing.Type[Bet] | tuple[typing.Type[Bet], ...]
    ):
        """
        Returns the bets a player has matching the type

        Notably, bets like Place(4, 1.0) will match to Place(6, 1.0).
        """
        return [x for x in self.bets if isinstance(x, bet_type)]

    def has_bets(self, bet_type: typing.Type[Bet] | tuple[typing.Type[Bet], ...]):
        return len(self.get_bets_by_type(bet_type)) > 0

    def remove_bet(self, bet: Bet) -> None:
        if bet in self.bets and bet.is_removable(self.table):
            self.bankroll += bet.amount
            self.bets.remove(bet)

    def add_strategy_bets(self) -> None:
        """Implement the given betting strategy"""
        if self.strategy is not None:
            self.strategy.update_bets(self)

    def update_bet(self, verbose: bool = False) -> None:
        for bet in self.bets[:]:
            result: BetResult = bet.get_result(self.table)
            self.bankroll += result.bankroll_change

            if verbose:
                self.print_bet_update(bet, result)

            if result.remove:
                self.bets.remove(bet)

    def print_bet_update(self, bet: Bet, result: BetResult) -> None:
        if result.won:
            print(f"{self.name} won ${result.amount - bet.amount} on {bet}!")
        elif result.lost:
            print(f"{self.name} lost ${bet.amount} on {bet}.")



from crapssim import Table
from crapssim.bet import PassLine
from crapssim.strategy import BetPassLine


def test_default_strategy():
    table = Table()
    table.add_player()
    assert table.players[0].strategy == BetPassLine(5)


def test_irremovable_bet():
    bet = PassLine(50)
    table = Table()
    table.add_player(500)
    table.fixed_run([(2, 2)])
    assert bet.is_removable(table) is False


def test_existing_bet():
    table = Table()
    table.add_player()
    bet_one = PassLine(50)
    table.players[0].add_bet(bet_one)
    bet_two = PassLine(50)
    table.players[0].add_bet(bet_two)

    bet_count = len(table.players[0].bets)
    bet_amount = table.players[0].bets[0].amount
    bankroll = table.players[0].bankroll
    total_bet_amount = table.players[0].total_bet_amount

    assert (bet_count, bet_amount, bankroll, total_bet_amount) == (1, 100, 0, 100)



"""Core strategies are strategies that can be either subclassed, or initialized to create other
strategies with the intended usage. Each of the strategies included in this package are intended
to be used as building blocks when creating strategies."""

import copy
import typing
from abc import ABC, abstractmethod
from typing import Protocol

from crapssim.bet import Bet, HardWay, Hop, Place
from crapssim.dice import Dice
from crapssim.point import Point

__all__ = [
    "Strategy",
    "AggregateStrategy",
    "NullStrategy",
    "AddIfTrue",
    "RemoveIfTrue",
    "ReplaceIfTrue",
    "AddIfNotBet",
    "AddIfPointOff",
    "AddIfPointOn",
    "AddIfNewShooter",
    "CountStrategy",
    "RemoveIfPointOff",
    "RemoveByType",
    "WinProgression",
]


class Table(Protocol):
    """Table functionality needed for strategy module."""

    dice: Dice
    point: Point
    new_shooter: bool


class Player(Protocol):
    """Player functionality needed for strategy module."""

    table: Table
    bankroll: float
    bets: list[Bet]

    def add_bet(self, bet: Bet) -> None: ...

    def already_placed_bets(self, bet: Bet) -> list[Bet]: ...

    def already_placed(self, bet: Bet) -> bool: ...

    def get_bets_by_type(
        self, bet_type: typing.Type[Bet] | tuple[typing.Type[Bet], ...]
    ): ...

    def remove_bet(self, bet: Bet) -> None: ...


class Strategy(ABC):
    """A Strategy is assigned to a player and determines what bets the player
    is going to make, remove, or change.
    """

    def after_roll(self, player: Player) -> None:
        """Method that can update the Strategy from the table/player after the dice are rolled but
        before the bets and the table are updated. For example, if you wanted to know whether the
        point changed from on to off you could do self.point_lost = table.point.status = "On" and
        table.dice.roll.total == 7. You couldn't do this in update_bets since the table has already
        been updated setting the point's status to Off.

        Parameters
        ----------
        player
            The Player to check for bets, etc.
        """

    @abstractmethod
    def completed(self, player: Player) -> bool:
        """If True, the Strategy is completed and the Player stops playing. If False, the Player
        keeps playing the Strategy."""

    @abstractmethod
    def update_bets(self, player: Player) -> None:
        """Add, remove, or change the bets on the table.

        This method is applied after the dice are rolled,
        the bets are updated, and the table is updated."""

    def __add__(self, other: "Strategy") -> "AggregateStrategy":
        return AggregateStrategy(self, other)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Strategy):
            return self.__class__ == other.__class__
        return NotImplemented

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"


class AggregateStrategy(Strategy):
    """A combination of multiple strategies."""

    def __init__(self, *strategies: Strategy):
        """A combination of multiple strategies. Strategies are applied in the order that is given.

        Parameters
        ----------
        strategies
            The strategies to combine to make the new strategy.
        """
        self.strategies = strategies

    def update_bets(self, player: Player) -> None:
        """Go through each of the strategies and run its update_bets method if the strategy has
        not been completed.

        Parameters
        ----------
        player
            The player to update the bets for.
        """
        for strategy in self.strategies:
            if not strategy.completed(player):
                strategy.update_bets(player)

    def completed(self, player: Player) -> bool:
        """Returns True if all the strategies in the AggregateStrategy are completed.

        Parameters
        ----------
        player
            The Player to check the strategy for.

        Returns
        -------
        A boolean representing whether the given strategy

        """
        return all(x.completed(player) for x in self.strategies)

    def __repr__(self) -> str:
        repr_strategies = [repr(x) for x in self.strategies]
        return f'{" + ".join(repr_strategies)}'


class NullStrategy(Strategy):
    """Strategy that bets nothing."""

    def update_bets(self, player: Player) -> None:
        pass

    def completed(self, player: Player) -> bool:
        return False

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"


class AddIfTrue(Strategy):
    """Strategy that places a bet if a given key taking Player as a parameter is True."""

    def __init__(self, bet: Bet, key: typing.Callable[[Player], bool]):
        """The strategy will place the given bet if the given key is True.

        Parameters
        ----------
        bet
            The Bet to place if key is True.
        key
            Callable with parameters of player and table
            returning a boolean to decide whether to place the bet.
        """

        super().__init__()
        self.bet = bet
        self.key = key

    def update_bets(self, player: Player) -> None:
        """If the key is True add the bet to the player and table.

        Parameters
        ----------
        player
            The Player to add the bet for.
        """
        if self.key(player) and self.bet.is_allowed(player):
            player.add_bet(self.bet)

    def completed(self, player: Player) -> bool:
        """The strategy is completed when the player  can't make a bet because their bankroll is too
         low and the player doesn't have any bets left on the table.

        Parameters
        ----------
        player
            The player to check whether the

        Returns
        -------
        True if the Player can't continue the strategy, otherwise False.
        """
        return (
            self.bet.amount > player.bankroll
            and sum(x.amount for x in player.bets) == 0
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(bet={self.bet}, " f"key={self.key})"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Strategy):
            return isinstance(other, type(self)) and self.bet == other.bet
        raise NotImplementedError


class RemoveIfTrue(Strategy):
    """Strategy that removes all bets that are True for a given key. The key takes the Bet and the
    Player as parameters."""

    def __init__(self, key: typing.Callable[["Bet", Player], bool]):
        """The strategy will remove all bets that are true for the given key.

        Parameters
        ----------
        key
            Callable with parameters of bet and player return True if the bet should be removed
            otherwise returning False.
        """
        super().__init__()
        self.key = key

    def update_bets(self, player: Player) -> None:
        """For each of the players bets if the key is True remove the bet from the table.

        Parameters
        ----------
        player
            The Player to remove the bets for.
        """
        bets_to_remove = []
        for bet in player.bets:
            if self.key(bet, player):
                bets_to_remove.append(bet)
        for bet in bets_to_remove:
            player.remove_bet(bet)

    def completed(self, player: Player) -> bool:
        """The strategy is completed when the player doesn't have any bets left on the table.

        Parameters
        ----------
        player
            The player to check whether the

        Returns
        -------
        True if the Player can't continue the strategy, otherwise False.
        """
        return sum(x.amount for x in player.bets) == 0

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(key={self.key})"


class ReplaceIfTrue(Strategy):
    """Strategy that iterates through the bets on the table and if the given key is true, replaces
    the bet with the given bet."""

    def __init__(self, bet: Bet, key: typing.Callable[[Bet, Player], bool]):
        self.key = key
        self.bet = bet

    def update_bets(self, player: Player) -> None:
        """Iterate through each bet for the player and if the self.key(bet, player) is True, remove
        the bet and replace it with self.bet.

        Parameters
        ----------
        player
            The player to check the bets for.
        """
        for bet in player.bets:
            if self.key(bet, player):
                player.remove_bet(bet)
                player.add_bet(self.bet)

    def completed(self, player: Player) -> bool:
        """The strategy is completed when the player  can't make a bet because their bankroll is too
         low and the player doesn't have any bets left on the table.

        Parameters
        ----------
        player
            The player to check whether the

        Returns
        -------
        True if the Player can't continue the strategy, otherwise False.
        """
        return (
            self.bet.amount > player.bankroll
            and sum(x.amount for x in player.bets) == 0
        )


class AddIfNotBet(AddIfTrue):
    """Strategy that adds a bet if it isn't on the table for that player. Equivalent of
    AddIfTrue(bet, lambda p: bet not in p.bets)"""

    def __init__(self, bet: Bet):
        """The strategy adds the given bet object to the table if it is not already on the table.

        Parameters
        ----------
        bet
            The bet to add if it isn't already on the table.
        """
        super().__init__(bet, lambda p: bet not in p.bets)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(bet={self.bet})"


class AddIfPointOff(AddIfTrue):
    """Strategy that adds a bet if the table point is Off, and the Player doesn't have a bet on the
    table. Equivalent to AddIfTrue(bet, lambda p: p.table.point.status == "Off"
                                        and bet not in p.bets)"""

    def __init__(self, bet: Bet):
        """Adds the given bet if the table point is Off and the player doesn't have that bet on the
        table.

        Parameters
        ----------
        bet
            The bet to add if the point is Off.
        """
        super().__init__(
            bet, lambda p: p.table.point.status == "Off" and bet not in p.bets
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(bet={self.bet})"


class AddIfPointOn(AddIfTrue):
    """Strategy that adds a bet if the table point is On, and the Player doesn't have a bet on the
    table. Equivalent to AddIfTrue(bet, lambda p: p.table.point.status == "On"
                                        and bet not in p.bets)"""

    def __init__(self, bet: Bet):
        """Add a bet if the point is On.

        Parameters
        ----------
        bet
            The bet to add if the point is On.
        """
        super().__init__(
            bet, lambda p: p.table.point.status == "On" and bet not in p.bets
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(bet={self.bet})"


class AddIfNewShooter(AddIfTrue):
    """Strategy that adds a bet if there is a new shooter at the table, and the Player doesn't have a bet on the
    table. Equivalent to AddIfTrue(bet, lambda p: p.table.new_shooter and bet not in p.bets)
    """

    def __init__(self, bet: Bet):
        """Add a bet if the point is On.

        Parameters
        ----------
        bet
            The bet to add if the point is On.
        """
        super().__init__(bet, lambda p: p.table.new_shooter and bet not in p.bets)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(bet={self.bet})"


class CountStrategy(AddIfTrue):
    """Strategy that checks how many bets exist of a certain type. If the number of bets of that
    type is less than the given count, it places the bet (if the bet isn't already on the table.)
    """

    def __init__(
        self,
        bet_type: typing.Type[Bet] | tuple[typing.Type[Bet], ...],
        count: int,
        bet: Bet,
    ):
        """If there are less than count number of bets placed by player with a given bet_type, it
        adds the given bet.

        Parameters
        ----------
        bet_type
            The types of bets to count.
        count
            How many of the bets to check against.
        bet
            The bet to place if there are less than count bets of a given type.
        """
        self.bet_type = bet_type
        self.count = count

        super().__init__(bet, key=self.key)

    def key(self, player: Player) -> bool:
        """Return True if the player has less than count number of bets for a given type and the
        bet that is intended to be placed isn't already on the table.

        Parameters
        ----------
        player
            The player to count the bets for.

        Returns
        -------
        Returns True if the player has less than count number of bets for a given type and the
        bet that is intended to be placed isn't already on the table, otherwise returns False.
        """
        count_of_bets_with_type = len(player.get_bets_by_type(bet_type=self.bet_type))
        identical_bet_is_not_on_table = self.bet not in player.bets

        return count_of_bets_with_type < self.count and identical_bet_is_not_on_table

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(bet_type={self.bet_type}, count={self.count}, "
            f"bet={self.bet})"
        )


class RemoveIfPointOff(RemoveIfTrue):
    """Strategy that removes a bet if the table point is Off

    This will match bets based on type, and number for Place and Hardway bets.
    It will not consider bet amounts when matching."""

    def __init__(self, bet: Bet):
        if not any([isinstance(bet, x) for x in [Place, HardWay, Hop]]):
            key = (
                lambda b, p: isinstance(b, type(self.bet))
                and p.table.point.status == "Off"
            )

        if isinstance(bet, Place):
            key = (
                lambda b, p: isinstance(b, Place)
                and b.number == self.bet.number
                and p.table.point.status == "Off"
            )
        if isinstance(bet, HardWay):
            key = (
                lambda b, p: isinstance(b, HardWay)
                and b.number == self.bet.number
                and p.table.point.status == "Off"
            )
        if isinstance(bet, Hop):
            key = (
                lambda b, p: isinstance(b, Hop)
                and b.result == self.bet.result
                and p.table.point.status == "Off"
            )

        super().__init__(key)
        self.bet = bet

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(bet={self.bet})"


class RemoveByType(RemoveIfTrue):
    """Remove any bets that are of the given type(s)."""

    def __init__(self, bet_type: typing.Type[Bet] | tuple[typing.Type[Bet], ...]):
        super().__init__(lambda b, p: isinstance(b, bet_type))


class WinProgression(Strategy):
    """Strategy that every time a bet is won, moves to the next amount in the progression and
    places a Field bet for that amount."""

    def __init__(self, first_bet: Bet, multipliers: list[typing.SupportsFloat]) -> None:
        """Creates the given the progression.

        Parameters
        ----------
        first_bet
            The initial bet, including the starting amount
        progression
            A list of multipliers on the bet amounts to make. As you win, progresses farther up list.
        """
        self.bet = first_bet
        self.multipliers = multipliers
        self.current_progression = 0

    def completed(self, player: Player) -> bool:
        """If the players bankroll is below the minimum amount in the progression and if they
        have no more bets on the table the strategy is completed.

        Parameters
        ----------
        player
            The player to check the bankroll and bets for.

        Returns
        -------
        True if the
        """
        return (
            player.bankroll < min(float(x) for x in self.multipliers)
            and len(player.bets) == 0
        )

    def after_roll(self, player: Player) -> None:
        """If the field bet wins, increase the progression by 1, if it loses reset the progression
        to 0.

        Parameters
        ----------
        player
            The player to check the winning bets for.
        """

        win = all(x.get_result(player.table).won for x in player.bets)

        if win:
            self.current_progression += 1
        else:
            self.current_progression = 0

    def update_bets(self, player: Player) -> None:
        """If a bet isn't on the table, place one for the current progression amount.

        Parameters
        ----------
        player
            The player to place the bet for.
        """
        new_bet = copy.copy(self.bet)
        if self.current_progression >= len(self.multipliers):
            new_bet.amount = self.bet.amount * self.multipliers[-1]
        else:
            new_bet.amount = (
                self.bet.amount * self.multipliers[self.current_progression]
            )
        AddIfNotBet(new_bet).update_bets(player)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(first_bet={self.bet}, multipliers={self.multipliers})"



"""
Strategies to be assigned to players on the Craps table. The strategy determines what bets the
player should make, remove, or change. Strategies are applied for the player to change the bets
after the previous bets and table have been updated.
"""

from crapssim.strategy.odds import (
    ComeOddsMultiplier,
    DontComeOddsMultiplier,
    DontPassOddsMultiplier,
    OddsAmount,
    OddsMultiplier,
    PassLineOddsMultiplier,
)
from crapssim.strategy.single_bet import BetDontPass, BetPassLine, BetPlace
from crapssim.strategy.tools import (
    AddIfNewShooter,
    AddIfNotBet,
    AddIfPointOff,
    AddIfPointOn,
    AddIfTrue,
    AggregateStrategy,
    CountStrategy,
    RemoveIfPointOff,
    RemoveIfTrue,
    Strategy,
)

from . import examples, odds



import pytest

from crapssim import Table
from crapssim.bet import Come
from crapssim.point import Point
from crapssim.strategy import BetPassLine


def test_ensure_one_player():
    table = Table()
    count_zero = len(table.players)
    table.ensure_one_player()
    count_one = len(table.players)
    bankroll = table.players[0].bankroll
    strategy = table.players[0].strategy.__class__
    assert (count_zero, count_one, bankroll, strategy) == (0, 1, 100, BetPassLine)


def test_wrong_point_off():
    table = Table()
    table.add_player(bankroll=500)
    table.players[0].add_bet(Come(100))
    assert (len(table.players[0].bets), table.players[0].bankroll) == (0, 500)


def test_wrong_point_on():
    table = Table()
    table.point.number = 4
    table.add_player(bankroll=500)
    assert (len(table.players[0].bets), table.players[0].bankroll) == (0, 500)


@pytest.mark.parametrize(
    ["status", "number", "comparison"],
    [
        ("Off", None, "Off"),
        ("Off", None, "off"),
        ("On", 6, 6),
        ("On", 6, "6"),
        ("On", 8, "on"),
    ],
)
def test_point_equality(status, number, comparison):
    point = Point()
    point.number = number
    assert point == comparison


@pytest.mark.parametrize(["number", "comparison"], [(8, 6), (8, "6")])
def test_point_greater_than(number, comparison):
    point = Point()
    point.number = number
    assert point > comparison


@pytest.mark.parametrize(["number", "comparison"], [(4, 6), (4, "10")])
def test_point_less_than(number, comparison):
    point = Point()
    point.number = number
    assert point < comparison


@pytest.mark.parametrize("seed", [8, 15, 21234, 0])
def test_table_seed_idential(seed):
    table1 = Table(seed)
    table2 = Table(seed)

    table1.run(max_rolls=100)
    table2.run(max_rolls=100)

    assert table1.dice.result == table2.dice.result


@pytest.mark.parametrize(
    ["rolls", "n_shooters"],
    [
        ([], 1),
        ([(3, 4)], 1),
        ([(3, 3), (3, 4)], 2),
        ([(3, 3), (3, 4), (4, 4), (2, 6)], 2),
        ([(3, 3), (3, 4), (4, 4), (2, 6), (1, 1)], 2),
        ([(3, 3), (3, 4), (4, 4), (2, 6), (1, 1), (2, 2), (1, 6)], 3),
    ],
)
def test_table_n_shooters(rolls, n_shooters):
    table = Table()
    table.fixed_run(rolls)
    assert table.n_shooters == n_shooters


def test_table_rerunning_with_rolls():
    table = Table()
    table.run(max_rolls=10, verbose=False)

    assert table.dice.n_rolls == 10

    table.run(max_rolls=7, verbose=False)

    assert table.dice.n_rolls == 17


def test_table_rerunning_with_shooters():
    table = Table(seed=8)
    assert table.n_shooters == 1

    table.run(max_rolls=float("inf"), max_shooter=2)
    assert table.n_shooters == 2

    table.run(max_rolls=float("inf"), max_shooter=5)
    assert table.n_shooters == 7



import typing

from crapssim.bet import Bet, Come, DontCome, DontPass, Odds, PassLine
from crapssim.strategy.tools import Player, Strategy, Table


class OddsAmount(Strategy):
    """Strategy that takes places odds on a given number for a given bet type."""

    def __init__(
        self,
        base_type: typing.Type[PassLine | DontPass | Come | DontCome],
        odds_amounts: dict[int, typing.SupportsFloat],
    ):
        self.base_type = base_type
        self.odds_amounts = odds_amounts

    def completed(self, player: Player) -> bool:
        """Return True if there are no bets of base_type on the table.

        Parameters
        ----------
        player
            The player whose bets to check for.

        Returns
        -------
        True if there are no base type bets on the table, otherwise False.
        """
        return len([x for x in player.bets if isinstance(x, self.base_type)]) == 0

    def update_bets(self, player: Player) -> None:
        for number, amount in self.odds_amounts.items():
            bet = Odds(self.base_type, number, float(amount))
            if bet.is_allowed(player) and not player.already_placed(bet):
                player.add_bet(bet)


class PassLineOddsAmount(OddsAmount):
    def __init__(
        self,
        bet_amount: typing.SupportsFloat,
        numbers: tuple[int] = (4, 5, 6, 8, 9, 10),
    ):
        super().__init__(PassLine, {x: bet_amount for x in numbers})


class DontPassOddsAmount(OddsAmount):
    def __init__(
        self,
        bet_amount: typing.SupportsFloat,
        numbers: tuple[int] = (4, 5, 6, 8, 9, 10),
    ):
        super().__init__(DontPass, {x: bet_amount for x in numbers})


class ComeOddsAmount(OddsAmount):
    def __init__(
        self,
        bet_amount: typing.SupportsFloat,
        numbers: tuple[int] = (4, 5, 6, 8, 9, 10),
    ):
        super().__init__(DontPass, {x: bet_amount for x in numbers})


class DontComeOddsAmount(OddsAmount):
    def __init__(
        self,
        bet_amount: typing.SupportsFloat,
        numbers: tuple[int] = (4, 5, 6, 8, 9, 10),
    ):
        super().__init__(DontCome, {x: bet_amount for x in numbers})


class OddsMultiplier(Strategy):
    """Strategy that takes an AllowsOdds object and places Odds on it given either a multiplier,
    or a dictionary of points and multipliers."""

    def __init__(
        self,
        base_type: typing.Type[PassLine | DontPass | Come | DontCome],
        odds_multiplier: dict[int, int] | int,
    ):
        """Takes an AllowsOdds item (ex. PassLine, Come, DontPass) and adds a BaseOdds bet
        (either Odds or LayOdds) based on the odds_multiplier given.

        Parameters
        ----------
        base_type
            The bet that odds will be added to.
        odds_multiplier
            If odds_multiplier is an integer adds multiplier * base_bets amount to the odds.
            If the odds multiplier is a dictionary of integers, looks at the dictionary to
            determine what odds multiplier to use depending on the given point.
        """
        self.base_type = base_type

        if isinstance(odds_multiplier, int):
            self.odds_multiplier = {x: odds_multiplier for x in (4, 5, 6, 8, 9, 10)}
        else:
            self.odds_multiplier = odds_multiplier

    @staticmethod
    def get_point_number(bet: Bet, table: "Table"):
        if isinstance(bet, (PassLine, DontPass)):
            return table.point.number
        elif isinstance(bet, (Come, DontCome)):
            return bet.number
        else:
            raise NotImplementedError

    def update_bets(self, player: Player) -> None:
        """Add an Odds bet to the given base_types in the amount determined by the odds_multiplier.

        Parameters
        ----------
        player
            The player to add the odds bet to.
        """
        for bet in [x for x in player.bets if isinstance(x, self.base_type)]:
            point = self.get_point_number(bet, player.table)

            if point in self.odds_multiplier:
                multiplier = self.odds_multiplier[point]
            else:
                return

            amount = bet.amount * multiplier
            OddsAmount(self.base_type, {point: amount}).update_bets(player)

    def completed(self, player: Player) -> bool:
        """Return True if there are no bets of base_type on the table.

        Parameters
        ----------
        player
            The player whose bets to check for.

        Returns
        -------
        True if there are no base type bets on the table, otherwise False.
        """
        return len([x for x in player.bets if isinstance(x, self.base_type)]) == 0

    def get_odds_multiplier_repr(self) -> int | dict[int, int]:
        """If the odds_multiplier has multiple values return a dictionary with the values,
        if all the multipliers are the same return an integer of the multiplier."""
        if all([x == self.odds_multiplier[4] for x in self.odds_multiplier.values()]):
            odds_multiplier: int | dict[int, int] = self.odds_multiplier[4]
        else:
            odds_multiplier = self.odds_multiplier
        return odds_multiplier

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(base_type={self.base_type}, "
            f"odds_multiplier={self.get_odds_multiplier_repr()})"
        )


class PassLineOddsMultiplier(OddsMultiplier):
    """Strategy that adds an Odds bet to the PassLine bet. Equivalent to
    OddsMultiplier(PassLine, odds)."""

    def __init__(self, odds_multiplier: dict[int, int] | int | None = None):
        """Add odds to PassLine bets with the multiplier specified by the odds_multiplier variable.

        Parameters
        ----------
        odds_multiplier
            If odds_multiplier is an integer the bet amount is the PassLine bet amount *
            odds_multiplier.  If it's a dictionary it uses the PassLine bet's point to determine
            the multiplier. Defaults to {4: 3, 5: 4, 6: 5, 8: 5, 9: 4, 10: 3} which are 345x odds.
        """
        if odds_multiplier is None:
            odds_multiplier = {4: 3, 5: 4, 6: 5, 8: 5, 9: 4, 10: 3}
        super().__init__(PassLine, odds_multiplier)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(odds_multiplier={self.get_odds_multiplier_repr()})"


class DontPassOddsMultiplier(OddsMultiplier):
    """Strategy that adds a LayOdds bet to the DontPass bet. Equivalent to
    OddsMultiplier(DontPass, odds)"""

    def __init__(self, odds_multiplier: dict[int, int] | int | None = None):
        """Add odds to DontPass bets with the multiplier specified by odds.

        Parameters
        ----------
        odds_multiplier
            If odds_multiplier is an integer the bet amount is the DontPass bet amount *
            odds_multiplier. If it's a dictionary it uses the DontPass bet's point to determine the
            multiplier. Defaults to 6.
        """
        if odds_multiplier is None:
            odds_multiplier = 6
        super().__init__(DontPass, odds_multiplier)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(odds_multiplier={self.get_odds_multiplier_repr()})"


class ComeOddsMultiplier(OddsMultiplier):
    """Strategy that adds an Odds bet to the Come bet. Equivalent to
    OddsMultiplier(Come, odds)."""

    def __init__(self, odds_multiplier: dict[int, int] | int | None = None):
        """Add odds to Come bets with the multiplier specified by the odds_multiplier variable.

        Parameters
        ----------
        odds_multiplier
            If odds_multiplier is an integer the bet amount is the Come bet amount *
            odds_multiplier.  If it's a dictionary it uses the Come bet's point to determine
            the multiplier. Defaults to {4: 3, 5: 4, 6: 5, 8: 5, 9: 4, 10: 3} which are 345x odds.
        """
        if odds_multiplier is None:
            odds_multiplier = {4: 3, 5: 4, 6: 5, 8: 5, 9: 4, 10: 3}
        super().__init__(Come, odds_multiplier)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(odds_multiplier={self.get_odds_multiplier_repr()})"


class DontComeOddsMultiplier(OddsMultiplier):
    """Strategy that adds an Odds bet to the DontCome bet. Equivalent to
    OddsMultiplier(DontCome, odds)."""

    def __init__(self, odds_multiplier: dict[int, int] | int | None = None):
        """Add odds to DontCome bets with the multiplier specified by the odds_multiplier variable.

        Parameters
        ----------
        odds_multiplier
            If odds_multiplier is an integer the bet amount is the Come bet amount *
            odds_multiplier.  If it's a dictionary it uses the Come bet's point to determine
            the multiplier. Defaults to 6.
        """
        if odds_multiplier is None:
            odds_multiplier = 6
        super().__init__(DontCome, odds_multiplier)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(odds_multiplier={self.get_odds_multiplier_repr()})"
